/*
  Copyright (C) 2002-2024 CERN for the benefit of the ATLAS collaboration
*/

#ifndef OUTPUTSTREAMSEQUENCERSVC_H
#define OUTPUTSTREAMSEQUENCERSVC_H

/** @file OutputStreamSequencerSvc.h
 *  @brief This file contains the class definition for the OutputStreamSequencerSvc class.
 *  @author Peter van Gemmeren <gemmeren@anl.gov>
 **/

#include "GaudiKernel/ServiceHandle.h"
#include "Gaudi/Property.h"  // no forward decl: typedef
#include "GaudiKernel/IIncidentListener.h"
#include "AthenaBaseComps/AthService.h"

#include <memory>
#include <map>
#include <mutex>

// Forward declarations
class MetaDataSvc;


/** @class OutputStreamSequencerSvc
 *  @brief This class provides configuration properties to enable OutputStream file sequences
 **/
class OutputStreamSequencerSvc : public extends<AthService,
                                                IIncidentListener> {

 public:
  using IService::interfaceID;
  typedef std::pair<std::string,std::string> RangeReport_t; // (RangeID,RangeFileName)
  typedef std::unique_ptr<RangeReport_t>     RangeReport_ptr;

public: // Constructor and Destructor

   /// Standard Service Constructor
   OutputStreamSequencerSvc(const std::string& name, ISvcLocator* pSvcLocator);
   /// Destructor
   virtual ~OutputStreamSequencerSvc();

public: // Non-static members
   /// Required of all Gaudi services:
   virtual StatusCode initialize() override final;
   /// Required of all Gaudi services:
   virtual StatusCode finalize() override final;

   /// Incident service handle
   virtual void handle(const Incident& /*inc*/) override final;

   /// Returns sequenced file name for output stream
   std::string buildSequenceFileName(const std::string&);
   void publishRangeReport(const std::string& outputFile);
   RangeReport_ptr getRangeReport();

   /// The name of the incident that starts a new event sequence
   std::string  incidentName() const            { return m_incidentName.value(); }

   /// The current Event Range ID (only one range is returned)
   std::string  currentRangeID() const;

   /// set the RangeID (possibly temporarily) so the right Range Filename may be generated
   std::string  setRangeID(const std::string & rangeID);

   /// Is the service in active use? (true after the first range incident is handled)
   bool         inUse() const;
  
   /// Are there concurrent events? (threads>1)
   static bool  inConcurrentEventsMode();

   /// Last incident type that was handled
   const std::string& lastIncident()            { return m_lastIncident; }

private: // data
   ServiceHandle<MetaDataSvc> m_metaDataSvc;

   /// The event sequence number
   int m_fileSequenceNumber;

   /// Current EventRange ID constructed on the last NextRange incident
   std::string  m_currentRangeID;

   /// Recently constructed full file name (useful in single threaded processing)
   std::string  m_lastFileName;

   /// Last incident type that was handled
   std::string  m_lastIncident;

   /// EventRange ID for all slots
   std::vector<std::string>   m_rangeIDinSlot;

private: // properties
   /// SequenceIncidentName, incident name for triggering file sequencing.
   StringProperty             m_incidentName {this, "SequenceIncidentName", "",
         "Name of the incident that signals the next Event Range start" };
   /// Flag to switch on storage of reporting info in fnToRangeId
   BooleanProperty            m_reportingOn {this, "ReportingOn", false,
         "If True, keep info about Ranges for getRangeReport() calls"};

   std::map<std::string,std::string> m_fnToRangeId;
   std::map<std::string,std::string>::iterator m_finishedRange;

   mutable std::mutex         m_mutex;
};

#endif
