/*
  Copyright (C) 2002-2024 CERN for the benefit of the ATLAS collaboration
*/

#ifndef IOVSVC_CBNODE_H
#define IOVSVC_CBNODE_H

/*****************************************************************************
 *
 *  CBNode.h
 *  IOVSvc
 *
 *  Author: Charles Leggett
 *
 *  Tree node structure for callback function trigger tree
 *
 *****************************************************************************/

#include <set>
#include <string>

#include "AthenaKernel/IOVSvcDefs.h"
#include "SGTools/CallBackID.h"
#include <atomic>

namespace SG {
  class DataProxy;
}

class CBNode {
public:

  typedef IOVSvcCallBackFcn BFCN;

  class nodeOrder {
  public:
    bool operator() ( const CBNode &n1, const CBNode &n2 ) const {
      if ( n1.level() != n2.level() ) {
        return ( n1.level() < n2.level() );
      } else {
        return ( n1.serial() < n2.serial() );
      }
    }

    bool operator() ( const CBNode *n1, const CBNode *n2 ) const {
      if ( n1->level() != n2->level() ) {
        return ( n1->level() < n2->level() );
      } else {
        return ( n1->serial() < n2->serial() );
      }
    }
  };

  typedef std::set<CBNode*, CBNode::nodeOrder> nodeSet;
  typedef std::vector<const CBNode*> constNodeVec;

  CBNode() = delete;
  CBNode(const std::string& name, CBNode* parent);
  CBNode(const SG::DataProxy* proxy, const std::string& name, CBNode* parent);
  CBNode(BFCN* fcn, const CallBackID& cb, CBNode* parent);
  ~CBNode();
  
  void addParent(CBNode* parent);
  void addChild(CBNode* child);

  bool delParent(CBNode* parent);
  bool delChild(CBNode* child);

  const nodeSet& children() { return m_children; }
  const nodeSet& parents() { return m_parents; }
  constNodeVec children() const { return constNodeVec (m_children.begin(), m_children.end()); }
  constNodeVec parents() const { return constNodeVec (m_parents.begin(), m_parents.end()); }

  const std::string& name() const { return m_name; }
  int level() const { return m_level; }
  void setLevel(int l) { m_level = l; }

  bool trigger() const { return m_trig; }
  void setTrigger(bool b) { m_trig = b; }

  bool flag() const { return m_flag; }
  void setFlag(bool b) { m_flag = b; }

  const SG::DataProxy* proxy() const { return m_proxy; }
  void setProxy(const SG::DataProxy* p);

  BFCN* fcn() { return m_fcn; }
  const BFCN* fcn() const { return m_fcn; }

  unsigned int serial() const { return m_serial; }

private:
  nodeSet m_parents;
  nodeSet m_children;

  std::string m_name;
  int m_level;

  const SG::DataProxy* m_proxy;
  BFCN* m_fcn;
  CallBackID m_cbid;

  bool m_trig;
  bool m_flag;

  unsigned int m_serial;
  static std::atomic<unsigned int> s_serial;
};

#endif
