/*
  Copyright (C) 2002-2025 CERN for the benefit of the ATLAS collaboration
*/

// GeneratorFilters/DecaysFinalStateFilter
//
// Picks events with a given number of quarks, leptons and neutrinos from
// decays of a list of specified resonances (e.g. W, Z, ...).
// Will work only if resonances are explicitly included in event record.
//
// Examples:
//
//   topAlg.DecaysFinalStateFilter.PDGAllowedParents = [ 23 ]
//   topAlg.DecaysFinalStateFilter.NQuarks = 2
//   topAlg.DecaysFinalStateFilter.NChargedLeptons = 2
//   -> picks semileptonic ZZ decays
//
//   topAlg.DecaysFinalStateFilter.PDGAllowedParents = [ 23, 24, -24 ]
//   topAlg.DecaysFinalStateFilter.NQuarks = 2
//   topAlg.DecaysFinalStateFilter.NChargedLeptons = 2
//   -> allows W(qq)Z(ll) and Z(qq)Z(ll)
//
// Requirements can be put on each lepton flavor individually or on charged
// leptons inclusively. To put requirements on a single lepton flavor, set
// the corresponding property while leaving the properties for the other
// lepton flavors as well as ChargedLeptons as their default values.
//
// Authors:
// Kerim Suruliz Nov 2014
// Frank Siegert Nov 2014
// Jason Veatch Nov 2023

#include "GeneratorFilters/xAODDecaysFinalStateFilter.h"
//#include "GaudiKernel/MsgStream.h"
#include "TruthUtils/HepMCHelpers.h"
#include <cmath>


StatusCode xAODDecaysFinalStateFilter::filterInitialize() {
    CHECK(m_truthPartContKey.initialize());
    return StatusCode::SUCCESS;
}


StatusCode xAODDecaysFinalStateFilter::filterEvent() {
    int nElectrons = 0;
    int nMuons = 0;
    int nTaus = 0;
    int nChargedLeptons = 0;
    int nQuarks = 0;
    int nNeutrinos = 0;
    int nPhotons = 0;

    // Retrieve TruthGen container from xAOD Gen slimmer, contains all particles witout barcode_zero and
    // duplicated barcode ones
    SG::ReadHandle<xAOD::TruthParticleContainer> xTruthParticleContainer{m_truthPartContKey};
    CHECK(xTruthParticleContainer.isValid());


    // Loop over all particles in the event and build up the grid
    for (const xAOD::TruthParticle* part : *xTruthParticleContainer) {
            // look only at the allowed parents (e.g. W, Z)
            bool allowedParent = false;
            for (size_t i=0; i<m_PDGAllowedParents.size(); ++i) {
            if ( part->pdgId() == m_PDGAllowedParents[i]) allowedParent = true;
            }
            if (!allowedParent) continue;

            if (!part->decayVtx()) continue;
            const xAOD::TruthVertex* decayVertex = part->decayVtx();
            int num_outgoing_particles = decayVertex->nOutgoingParticles();

            for (int iOutPart = 0; iOutPart< num_outgoing_particles; iOutPart++) {
                const xAOD::TruthParticle* out_part =  decayVertex->outgoingParticle(iOutPart);
                const int apid = std::abs(out_part->pdgId());
                if ( apid <=  MC::BQUARK ) nQuarks++;
                else if (MC::isElectron(apid)) { nElectrons++; nChargedLeptons++; }
                else if (MC::isMuon(apid)) { nMuons++; nChargedLeptons++; }
                else if (MC::isTau(apid)) { nTaus++; nChargedLeptons++; }
                else if (MC::isSMNeutrino(apid)) nNeutrinos++;
                else if (MC::isPhoton(apid)) nPhotons++;
            }

       }//loop over TruthParticles
    

  if (nQuarks < m_MinNQuarks || (m_NQuarks != -1 && nQuarks != m_NQuarks)) {
    setFilterPassed(false);
  }
  else if (nElectrons < m_MinNElectrons || (m_NElectrons != -1 && nElectrons != m_NElectrons)) {
    setFilterPassed(false);
  } 
  else if (nMuons < m_MinNMuons || (m_NMuons != -1 && nMuons != m_NMuons)) {
    setFilterPassed(false);
  } 
  else if (nTaus < m_MinNTaus || (m_NTaus != -1 && nTaus != m_NTaus)) {
    setFilterPassed(false);
  } 
  else if (nChargedLeptons < m_MinNChargedLeptons || (m_NChargedLeptons != -1 && nChargedLeptons != m_NChargedLeptons)) {
    setFilterPassed(false);
  } 
  else if (nNeutrinos < m_MinNNeutrinos || (m_NNeutrinos != -1 && nNeutrinos != m_NNeutrinos)) {
    setFilterPassed(false);
  } 
  else if (nPhotons < m_MinNPhotons || (m_NPhotons != -1 && nPhotons != m_NPhotons)) {
    setFilterPassed(false);
  } 
  else {
    setFilterPassed(true);
  }
  return StatusCode::SUCCESS;
}
