/*
  Copyright (C) 2002-2024 CERN for the benefit of the ATLAS collaboration
*/

/**
 * @file PixelClusterization.h
 * (c) ATLAS Detector software
 * @author D.Calvet, N.Nesvadba, T.Lari 
 * (modified s.roe, Dec 2008)
 **/

#ifndef INDETRIOMAKER_PIXELCLUSTERIZATION_H
#define INDETRIOMAKER_PIXELCLUSTERIZATION_H
//STL
#include <string>

//Tools
#include "AthenaMonitoringKernel/Monitored.h"
#include "AthenaMonitoringKernel/GenericMonitoringTool.h"

//Gaudi
#include "GaudiKernel/ToolHandle.h"
#include "GaudiKernel/ServiceHandle.h"
// Base class
#include "AthenaBaseComps/AthReentrantAlgorithm.h"
#include "StoreGate/ReadHandleKey.h"
#include "StoreGate/WriteHandleKey.h"

//InDet
//can't fwd declare this, needed for typedef to Pixel_RDO_Container
#include "InDetPrepRawData/SiClusterContainer.h"
#include "InDetPrepRawData/PixelClusterContainer.h"
#include "SiClusterizationTool/PixelGangedAmbiguitiesFinder.h"
#include "InDetRawData/InDetRawDataCollection.h"
#include "InDetRawData/PixelRDO_Container.h"
#include "TrigSteeringEvent/TrigRoiDescriptorCollection.h"
#include "IRegionSelector/IRegSelTool.h"
#include "SiClusterizationTool/IPixelClusteringTool.h"


// Fwd declarations
class PixelRDORawData;
class PixelID;

namespace InDet {
  class PixelGangedAmbiguitiesFinder;
}


namespace InDet {
/**
 * Top Algorithm to clusterize the pixel RDOs
 * (InDetRawDataCollection<PixelRDORawData> -> RecInputCollection)
 * The clustering algorithm is actually a private Tool in the
 * SiClusterizationTool package
 **/
class PixelClusterization : public AthReentrantAlgorithm {
public:
  typedef InDetRawDataCollection<PixelRDORawData> COLLECTION;

  /// Constructor with parameters:
  PixelClusterization(const std::string &name,ISvcLocator *pSvcLocator);

  //@name Usual algorithm methods 
  //@{
  virtual StatusCode initialize() override;
  virtual StatusCode execute(const EventContext& ctx) const override;
  /**    @name Disallow default instantiation, copy, assignment */
  //@{
  //@}
  PixelClusterization() = delete;
  PixelClusterization(const PixelClusterization&) = delete;
  PixelClusterization &operator=(const PixelClusterization&) = delete;
   //@}
                              
private:
  SG::ReadHandleKey<PixelRDO_Container> m_rdoContainerKey{
    this, "DataObjectName", "PixelRDOs"};
  SG::ReadHandleKey<TrigRoiDescriptorCollection> m_roiCollectionKey{
    this, "RoIs", "", "RoIs to read in"};

  ToolHandle<IPixelClusteringTool> m_clusteringTool{
    this, "clusteringTool", "InDet::MergedPixelsTool"};
  /// class to find out which clusters shares ganged pixels
  ToolHandle<PixelGangedAmbiguitiesFinder> m_gangedAmbiguitiesFinder{
    this, "gangedAmbiguitiesFinder", "InDet::PixelGangedAmbiguitiesFinder"};
  ToolHandle<IRegSelTool> m_regionSelector{
    this, "RegSelTool", "RegSelTool/RegSelTool_Pixel" };
  ToolHandle<GenericMonitoringTool> m_monTool{
    this, "MonTool", "", "Monitoring tool" };

  BooleanProperty m_roiSeeded{this, "isRoI_Seeded", false, "Use RoI"};
  const PixelID* m_idHelper;

  SG::WriteHandleKey<PixelClusterContainer> m_clusterContainerKey{
    this, "ClustersName", "PixelClusters", "Pixel cluster container"};
  SG::WriteHandleKey<SiClusterContainer> m_clusterContainerLinkKey{
    this, "ClustersLinkName_", "PixelClusters" ,
    "Pixel cluster container link name (don't set this)"};
  SG::WriteHandleKey<PixelGangedClusterAmbiguities> m_ambiguitiesMapKey{
    this, "AmbiguitiesMap", "PixelClusterAmbiguitiesMap",
    "Ambiguity Map container"};

  SG::UpdateHandleKey<InDet::PixelClusterContainerCache> m_clusterContainercacheKey{
    this, "ClusterContainerCacheKey", "",
    "Optional External Pixel cluster Cache"};
  BooleanProperty m_useDataPoolWithCache{
    this, "useDataPoolWithCache", false, "use DataPool With Cache"};
};

}//end of ns

#endif // INDETRIOMAKER_PIXELCLUSTERIZATION_H
