/*
  Copyright (C) 2002-2025 CERN for the benefit of the ATLAS collaboration
*/

#ifndef JIVEXML_XAODJETRETRIEVER_H
#define JIVEXML_XAODJETRETRIEVER_H

#include <string>
#include <vector>
#include <map>

#include "JiveXML/IDataRetriever.h"
#include "AthenaBaseComps/AthAlgTool.h"
#include "xAODJet/JetContainer.h"
#include "AsgTools/AnaToolHandle.h"
#include "FTagAnalysisInterfaces/IBTaggingSelectionTool.h"


namespace JiveXML{

  /**
   * @class xAODJetRetriever
   * @brief Retrieves all @c Jet @c objects (JetAODCollection etc.)
   *
   *  - @b Properties
   *    - PriorityJetCollection: First collection to be retrieved, displayed
   *      in Atlantis without switching. All other collections are
   *      also retrieved.
   *    - OtherJetCollections
   *    - DoWriteHLT
   *    - DoWriteAllCollections
   *    - WriteJetQuality
   *    - TracksName
   *    - CDIPaths
   *
   *  - @b Retrieved @b Data
   *    - Usual four-vectors: phi, eta, et etc.
   *    - Associations for clusters and tracks via ElementLink: key/index scheme
   */
  class xAODJetRetriever : virtual public IDataRetriever,
			   public AthAlgTool {

  public:

    /// Standard Constructor
    xAODJetRetriever(const std::string& type,const std::string& name,const IInterface* parent);

    virtual StatusCode retrieve(ToolHandle<IFormatTool> &FormatTool);
    /// Puts the variables into a DataMap
    const DataMap getData(const xAOD::JetContainer*, const std::string &jetkey);
    /// Gets the StoreGate keys for the desired containers
    const std::vector<std::string> getKeys();
    /// Return the name of the data type that is generated by this retriever
    virtual std::string dataTypeName() const { return m_typeName; };

    virtual StatusCode initialize();

  private:

    /// The data type that is generated by this retriever
    const std::string m_typeName = "Jet";

    Gaudi::Property<bool> m_doWriteHLT {this, "DoWriteHLT", false, "Write out other collections that have HLT in the name"};
    Gaudi::Property<bool> m_doWriteAllCollections {this, "DoWriteAllCollections", false, "Write out all Jet collections"};


    Gaudi::Property<std::vector<std::string>> m_otherKeys {this, "OtherJetCollections", {"AntiKt4EMTopoJets", "AntiKt4LCTopoJets", "AntiKt10LCTopoJets", "AntiKt10UFOCSSKJets"}, "Other collections to be retrieved. If DoWriteAllCollections is set to true all available Jet collections will be retrieved"};
    Gaudi::Property<std::string> m_priorityKey {this,"PriorityJetCollection","AntiKt4TopoEMJets", "Name of the priority Jet container that will be written out first"};
    Gaudi::Property<bool> m_writeJetQuality{
      this, "WriteJetQuality", false, "Write out extended jet quality information. False by default."};
    Gaudi::Property<std::string> m_tracksName{
      this, "TracksName", "InDetTrackParticles_xAOD", "Name of the track container to be retrieved"};
    Gaudi::Property<std::vector<std::string>> m_bTaggerNames{
      this, "BTaggerNames", {"DL1dv01", "GN2v01"}, "Names of the b-taggers to be retrieved"};
    Gaudi::Property<std::vector<std::string>> m_CDIPaths{
      this, "CDIPaths", {"xAODBTaggingEfficiency/13p6TeV/2023-22-13p6TeV-MC21-CDI_Test_2023-08-1_v1.root", "xAODBTaggingEfficiency/13p6TeV/2023-02_MC23_CDI_GN2v01-noSF.root"}, "Paths to the CDI files storing the b-tagger properties"};

    std::unordered_map<std::string, asg::AnaToolHandle<IBTaggingSelectionTool>> m_btagSelTools;
    unsigned int m_nTaggers=0;

  };
}
#endif
