/*
  Copyright (C) 2002-2025 CERN for the benefit of the ATLAS collaboration
*/

/*
 * Implementation file that mainly contains the code logic
 * dealing with Truth - record classification
 * Contributors:  Pierre-Antoine Delsart
 *                Andrii Verbytskyi <andrii.verbytskyi@mpp.mpg.de>
 */

#include "MCTruthClassifier/MCTruthClassifier.h"
#include "AsgDataHandles/ReadHandle.h"
#include "TruthUtils/MagicNumbers.h"
#ifndef XAOD_ANALYSIS
#include "AtlasHepMC/GenEvent.h"
#include "AtlasHepMC/GenVertex.h"
#include "AtlasHepMC/GenParticle.h"
#endif
#include "TruthUtils/DecayProducts.h"
using namespace MCTruthPartClassifier;
using std::abs;

#ifndef XAOD_ANALYSIS
std::pair<ParticleType, ParticleOrigin>
MCTruthClassifier::particleTruthClassifier(const HepMcParticleLink& theLink, MCTruthPartClassifier::Info* info /*= nullptr*/) const {
  // Retrieve the links between HepMC and xAOD::TruthParticle
  const EventContext& ctx = info ? info->eventContext : Gaudi::Hive::currentContext();
  SG::ReadHandle<xAODTruthParticleLinkVector> truthParticleLinkVecReadHandle(m_truthLinkVecReadHandleKey, ctx);
  if (!truthParticleLinkVecReadHandle.isValid()) {
    ATH_MSG_WARNING(" Invalid ReadHandle for xAODTruthParticleLinkVector with key: " << truthParticleLinkVecReadHandle.key());
    return std::make_pair(Unknown, NonDefined);
  }
  ElementLink<xAOD::TruthParticleContainer> tplink = truthParticleLinkVecReadHandle->find (theLink);
  if (tplink.isValid()) {
    return particleTruthClassifier (*tplink, info);
  }
  return std::make_pair(Unknown, NonDefined);
}

std::pair<ParticleType, ParticleOrigin>
MCTruthClassifier::particleTruthClassifier(HepMC::ConstGenParticlePtr theGenPart, MCTruthPartClassifier::Info* info /*= nullptr*/) const {
  ParticleType partType = Unknown;
  ParticleOrigin partOrig = NonDefined;

  if (!theGenPart) return std::make_pair(partType, partOrig);

  // Retrieve the links between HepMC and xAOD::TruthParticle
  const EventContext& ctx = info ? info->eventContext : Gaudi::Hive::currentContext();

  SG::ReadHandle<xAODTruthParticleLinkVector> truthParticleLinkVecReadHandle(m_truthLinkVecReadHandleKey, ctx);
  if (!truthParticleLinkVecReadHandle.isValid()) {
    ATH_MSG_WARNING( " Invalid ReadHandle for xAODTruthParticleLinkVector with key: " << truthParticleLinkVecReadHandle.key());
    return std::make_pair(partType, partOrig);
  }
  for (const auto *const entry : *truthParticleLinkVecReadHandle) {
    if (entry->first.isValid() && entry->second.isValid() && HepMC::is_same_particle(entry->first,theGenPart)) {
      const xAOD::TruthParticle* truthParticle = *entry->second;
      if (!theGenPart || !truthParticle ||
          theGenPart->pdg_id() != truthParticle->pdgId() ||
          theGenPart->status() != truthParticle->status() ||
          HepMC::is_same_particle(theGenPart,truthParticle)) {
        ATH_MSG_DEBUG(
                      "HepMC::GenParticle and xAOD::TruthParticle do not match");
        return std::make_pair(partType, partOrig);
      }
      return particleTruthClassifier(truthParticle, info);
    }
  }
  return std::make_pair(partType, partOrig);
}
#endif

std::pair<ParticleType, ParticleOrigin>
MCTruthClassifier::particleTruthClassifier(const xAOD::TruthParticle* thePart, MCTruthPartClassifier::Info* infoin /*= nullptr*/) const {
  MCTruthPartClassifier::Info tmpinfo;
  MCTruthPartClassifier::Info& info = (infoin) ? *infoin : tmpinfo;

  ATH_MSG_DEBUG("Executing particleTruthClassifier");

  ParticleType partType = Unknown;
  ParticleOrigin partOrig = NonDefined;
  if (!thePart) {
    return std::make_pair(partType, partOrig);
  }
  info.genPart = thePart;

  // retrieve collection and get a pointer
  SG::ReadHandle<xAOD::TruthParticleContainer> truthParticleContainerReadHandle(m_truthParticleContainerKey,info.eventContext);
  if (!truthParticleContainerReadHandle.isValid()) {
    ATH_MSG_WARNING( " Invalid ReadHandle for xAOD::TruthParticleContainer with key: " << truthParticleContainerReadHandle.key());
    return std::make_pair(partType, partOrig);
  }

  ATH_MSG_DEBUG("xAODTruthParticleContainer with key  " << truthParticleContainerReadHandle.key() << " has valid ReadHandle ");

  if (!MC::isStable(thePart) && !MC::isDecayed(thePart)) {
    return std::make_pair(GenParticle, partOrig);
  }
  const bool isPartHadr = MC::isHadron(thePart) && !MC::isBeam(thePart);
  if (MC::isDecayed(thePart) && (!MC::isTau(thePart) && !isPartHadr)) return std::make_pair(GenParticle, partOrig);

  // SUSY datasets: tau(status==2)->tau(status==2)
  if (MC::isDecayed(thePart) && MC::isTau(thePart)) {
    const xAOD::TruthVertex* endVert = thePart->decayVtx();
    if (endVert) {
      const int numberOfChildren = endVert->nOutgoingParticles();
      if (numberOfChildren == 1 && MC::isTau(endVert->outgoingParticle(0))) {
        return std::make_pair(GenParticle, partOrig);
      }
    }
  }

  if (MC::isStable(thePart) && MC::isSUSY(thePart)) return std::make_pair(SUSYParticle, partOrig);

  if (MC::isStable(thePart) && MC::isBSM(thePart)) return std::make_pair(OtherBSMParticle, partOrig);

  if (MC::isDecayed(thePart) &&
      (!MC::isElectron(thePart) && !MC::isMuon(thePart) &&
       !MC::isTau(thePart) && !MC::isPhoton(thePart)) &&
      !isPartHadr)
    return std::make_pair(GenParticle, partOrig);

  // FIXME vetoing protons here to preserve previous behaviour
  if (MC::isNucleus(thePart) && std::abs(thePart->pdgId()) != MC::PROTON) return std::make_pair(NuclFrag, partOrig);

  if ( !MC::isSMLepton(thePart) && !MC::isPhoton(thePart)  && !isPartHadr) return std::make_pair(partType, partOrig);
  // don't consider generator particles

  const xAOD::TruthVertex* partOrigVtx = thePart->hasProdVtx() ? thePart->prodVtx() : nullptr;

  const xAOD::TruthParticle* parent{};
  if (partOrigVtx) {
    for (const auto& temp: partOrigVtx->particles_in()) {if (temp)  parent = temp;}
  }
  const int parentPDG = parent?parent->pdg_id():0;
  info.setMotherProperties(parent);

  if (!partOrigVtx && HepMC::is_simulation_particle(thePart)) {
    return std::make_pair(NonPrimary, partOrig);
  }
  if (!partOrigVtx && MC::isElectron(thePart)) {
    // to define electron outcome status
    bool isPrompt = false; // updated by defOrigOfElectron
    partOrig = defOrigOfElectron(*truthParticleContainerReadHandle, thePart, isPrompt, info);
    return std::make_pair(UnknownElectron, partOrig);
  }
  if (!partOrigVtx && MC::isMuon(thePart)) {
    // to define electron outcome status
    bool isPrompt = false; // updated by defOrigOfMuon
    partOrig = defOrigOfMuon(*truthParticleContainerReadHandle, thePart, isPrompt, info);
    return std::make_pair(UnknownMuon, partOrig);
  }
  if (!partOrigVtx && MC::isTau(thePart)) {
    // to define electron outcome status
    partOrig = defOrigOfTau(*truthParticleContainerReadHandle, thePart, parentPDG, info);
    return std::make_pair(UnknownTau, partOrig);
  }
  if (!partOrigVtx && MC::isPhoton(thePart)) {
    // to define photon outcome
    bool isPrompt = false; // updated by defOrigOfPhoton
    partOrig = defOrigOfPhoton(*truthParticleContainerReadHandle, thePart, isPrompt, info);
    return std::make_pair(UnknownPhoton, partOrig);
  }
  if (!partOrigVtx && MC::isNeutrino(thePart)) {
    // to define neutrino outcome
    info.particleOutCome = NonInteract;
    return std::make_pair(Neutrino, partOrig);
  }

  if (thePart && info.Mother() && HepMC::is_same_generator_particle(thePart,info.Mother()))
    return std::make_pair(NonPrimary, partOrig);

  if (isPartHadr) return std::make_pair(Hadron, partOrig);

  if (partOrigVtx  && parentPDG == 0 && partOrigVtx->nOutgoingParticles() == 1 &&
      partOrigVtx->nIncomingParticles() == 0) {
    if (MC::isElectron(thePart)) {
      info.particleOutCome = defOutComeOfElectron(thePart);
      return std::make_pair(IsoElectron, SingleElec);
    }
    if (MC::isMuon(thePart)) {
      info.particleOutCome = defOutComeOfMuon(thePart);
      return std::make_pair(IsoMuon, SingleMuon);
    }
    if (MC::isTau(thePart)) {
      info.particleOutCome = defOutComeOfTau(thePart);
      return std::make_pair(IsoTau, SingleTau);
    }
    if (MC::isPhoton(thePart)) {
      info.particleOutCome = defOutComeOfPhoton(thePart);
      return std::make_pair(IsoPhoton, SinglePhot);
    }
  }

  if (parentPDG == thePart->pdg_id() &&  parent && parent->status() == 3 && MC::isDecayed(thePart)) return std::make_pair(GenParticle, partOrig);

  if (MC::isElectron(thePart)) {
    bool isPrompt = false; // updated by defOrigOfElectron
    partOrig = defOrigOfElectron(*truthParticleContainerReadHandle, thePart, isPrompt, info);
    partType = defTypeOfElectron(partOrig, isPrompt);
  } else if (MC::isMuon(thePart)) {
    bool isPrompt = false; // updated by defOrigOfMuon
    partOrig = defOrigOfMuon(*truthParticleContainerReadHandle, thePart, isPrompt, info);
    partType = defTypeOfMuon(partOrig, isPrompt);
  } else if (MC::isTau(thePart)) {
    partOrig = defOrigOfTau(*truthParticleContainerReadHandle, thePart, parentPDG, info);
    partType = defTypeOfTau(partOrig);
  } else if (MC::isPhoton(thePart)) {
    bool isPrompt = false; // updated by defOrigOfPhoton
    partOrig = defOrigOfPhoton(*truthParticleContainerReadHandle, thePart, isPrompt, info);
    partType = defTypeOfPhoton(partOrig);
  } else if (MC::isNeutrino(thePart)) {
    bool isPrompt = false; // updated by defOrigOfNeutrino
    partOrig = defOrigOfNeutrino(*truthParticleContainerReadHandle, thePart, isPrompt, info);
    partType = Neutrino;
  }

  ATH_MSG_DEBUG("particleTruthClassifier  succeeded ");
  return std::make_pair(partType, partOrig);
}


bool MCTruthClassifier::TruthLoopDetectionMethod1(const xAOD::TruthVertex * childOrigVtx, const xAOD::TruthParticle* parent) const
{
  // Start of method 1 of protecting against loops
  const int parentPDG = parent->pdgId();
  for (const auto& aChild: childOrigVtx->particles_out()) {
    if (!aChild)  continue;
    if (parentPDG == aChild->pdgId() && HepMC::is_same_generator_particle(aChild, parent)) {
      // One of the children produced in the decay of parent is
      // actually the same particle.  NB In the case of multiple
      // children this child may not necessarily be
      // thePriPart. Does this matter?
      return true;
    }
  }

  // to resolve Sherpa loop
  const xAOD::TruthVertex* parentOrigVtx = parent->hasProdVtx() ? parent->prodVtx() : nullptr;
  if (parentOrigVtx && HepMC::is_same_vertex(parentOrigVtx,childOrigVtx)) {
    // The "parent" and the "child" have the same production vertex.
    return true;
  }
  return false;
  // End of method 1 of protecting against loops
}

bool MCTruthClassifier::TruthLoopDetectionMethod2(const xAOD::TruthParticle* child, const xAOD::TruthParticle* parent) const
{
  // Start of method 2 of protecting against loops
  // to prevent Sherpa loop
  const xAOD::TruthVertex* child_prdVtx{};
  const xAOD::TruthVertex* child_endVtx{};
  if (child) {
    child_prdVtx = child->hasProdVtx() ? child->prodVtx() : nullptr;
    child_endVtx = child->decayVtx();
  }
  const xAOD::TruthVertex* parent_prdVtx{};
  const xAOD::TruthVertex* parent_endVtx{};
  if (parent) {
    parent_prdVtx = parent->hasProdVtx() ? parent->prodVtx() : nullptr;
    parent_endVtx = parent->decayVtx();
  }
  // V0->parent->V1-> ...->V2->child->V3
  // V3 == V0 && V1 == V2
  return (child_endVtx == parent_prdVtx && child_prdVtx == parent_endVtx);
}


ParticleOrigin MCTruthClassifier::defOrigOfElectron(const xAOD::TruthParticleContainer& xTruthParticleContainer,
                                                    const xAOD::TruthParticle* thePart,
                                                    bool& isPrompt,
                                                    MCTruthPartClassifier::Info& info) const
{
  ATH_MSG_DEBUG("Executing DefOrigOfElectron ");

  // Find the first copy of this particle stored in the xAOD::TruthParticleContainer (i.e. the particle prior to any interactions)
  const xAOD::TruthParticle* thePriPart = MC::findMatching(xTruthParticleContainer, thePart);
  if (!thePriPart) return NonDefined;
  if (!MC::isElectron(thePriPart)) return NonDefined;

  //-- to define electron outcome status
  info.particleOutCome = defOutComeOfElectron(thePriPart);

  const xAOD::TruthVertex* partOrigVtx = thePriPart->hasProdVtx() ? thePriPart->prodVtx() : nullptr;
  if (!partOrigVtx) return NonDefined;

  int numOfParents = partOrigVtx->nIncomingParticles();
  if (numOfParents > 1) ATH_MSG_DEBUG("DefOrigOfElectron:: electron has more than one parent.");

  const xAOD::TruthParticle* ancestor = MC::findMother(thePriPart);
  info.setMotherProperties(ancestor);
  if (!ancestor) { return NonDefined; } // After this point "ancestor" cannot be nullptr

  // Start of method 1 of protecting against loops
  bool samePart = TruthLoopDetectionMethod1(partOrigVtx, ancestor);
  // to resolve Sherpa loop
  // End of method 1 of protecting against loops

  if ((MC::isMuon(ancestor) || MC::isTau(ancestor) || MC::isW(ancestor)) && ancestor->hasProdVtx() && !samePart) {
    int pPDG(0);
    const xAOD::TruthParticle* AncestorParent{};
    do {
      pPDG = 0; // reset pPDG
      AncestorParent = MC::findMother(ancestor);
      // Start of method 2 of protecting against loops
      // to prevent Sherpa loop
      if (ancestor == AncestorParent) { break; }
      if (TruthLoopDetectionMethod2(ancestor,AncestorParent)) {
        AncestorParent = ancestor;
        break;
      }
      // End of method 2 of protecting against loops
      // FIXME why are slightly different criteria used in method 1 and method 2???
      if (AncestorParent) { pPDG = AncestorParent->pdgId(); } // Only set pPDG in the case that we aren't in a loop.
      if (MC::isMuon(pPDG) || MC::isTau(pPDG) || MC::isW(pPDG)) { // There will be another iteration so set ancestor to AncestorParent
        ancestor = AncestorParent; // AncestorParent is not nullptr here
      }
    } while ((MC::isMuon(pPDG) || MC::isTau(pPDG) || MC::isW(pPDG)));

    if (MC::isMuon(pPDG) || MC::isTau(pPDG) || MC::isW(pPDG) || MC::isZ(pPDG) || MC::isHiggs(pPDG) ||
        MC::isMSSMHiggs(pPDG) || MC::isHeavyBoson(pPDG) || MC::isTop(pPDG) || // MSSM Higgs bosons, Heavy bosons( Z', Z'', W'+)
        abs(pPDG) == MC::WBOSON_LRSM || MC::isNeutrinoRH(pPDG) || // Left-right symmetric model WBoson || Right-handed neutrino (Pythia-specific)
        MC::isSUSY(pPDG))
      ancestor = AncestorParent; // FIXME difference in behaviour compared to MCTruthClassifier::defOrigOfMuon/Neutrino
  }

  info.setMotherProperties(ancestor);
  const int ancestorPDG = ancestor->pdgId();
  const xAOD::TruthVertex* ancestorOrigVert = ancestor->hasProdVtx() ? ancestor->prodVtx() : nullptr;
  partOrigVtx = ancestor->decayVtx();
  numOfParents = partOrigVtx->nIncomingParticles();
  const int numberOfChildren = partOrigVtx->nOutgoingParticles();

  // Determine decay products
  auto DP = DecayProducts(partOrigVtx);
  const int NumOfPhot = DP.pd(MC::PHOTON);
  const int NumOfEl = DP.pd(MC::ELECTRON);
  const int NumOfPos = DP.pd(MC::POSITRON);
  const int NumOfquark = DP.apd({MC::DQUARK,MC::UQUARK,MC::SQUARK,MC::CQUARK,MC::BQUARK,MC::TQUARK});
  const int NumOfgluon = DP.apd(MC::GLUON);
  const int NumOfElNeut = DP.apd(MC::NU_E);
  const int NumOfLQ = DP.apd(MC::LEPTOQUARK);
  const int NumOfMuPl = DP.pd(-MC::MUON);
  const int NumOfMuMin = DP.pd(MC::MUON);
  const int NumOfMuNeut = DP.apd(MC::NU_MU);
  const int NumOfTau = DP.apd(MC::TAU);
  const int NumOfTauNeut = DP.apd(MC::NU_TAU);

  samePart = false;
  int NumOfNucFr(0);
  const bool possibleNuclearFragment = (numOfParents == 1 && (MC::isPhoton(ancestorPDG) || MC::isElectron(ancestorPDG) || abs(ancestorPDG) == MC::PIPLUS));
  for (const auto& aChild: partOrigVtx->particles_out()) {
    if (!aChild) continue;
    const int childPDG = aChild->pdgId();
    if (abs(childPDG) == abs(ancestorPDG) && HepMC::is_same_generator_particle(aChild, ancestor)) samePart = true;
    if (possibleNuclearFragment &&
        (MC::isNucleus(childPDG) || childPDG == 0 || childPDG == MC::PROTON || childPDG == MC::NEUTRON || // FIXME Do we really expect particles with PDG_ID = 0 in the truth record?
         abs(childPDG) == MC::PIPLUS || abs(childPDG) == MC::PI0))
      NumOfNucFr++;
  }
  // End of section determining decay products

  if (MC::isPhoton(ancestorPDG) && ancestorOrigVert) {
    if (ancestorOrigVert->nIncomingParticles() > 1) { ATH_MSG_DEBUG("DefOrigOfElectron:: photon has more than one parent.");  }
    for (const auto& photonParent: ancestorOrigVert->particles_in()) {
      if (!photonParent) continue;
      info.photonMother = photonParent; // FIXME Just taking the first valid particle...
    }
  }

  if ((MC::isPhoton(ancestorPDG) && numberOfChildren == 2 && NumOfEl == 1 && NumOfPos == 1) || (MC::isPhoton(ancestorPDG) && numberOfChildren == 1 && (NumOfEl == 1 || NumOfPos == 1))) return PhotonConv;

  // e,gamma,pi+Nuclear->NuclearFragments+nuclons+e
  if ((numOfParents == 1 && (MC::isPhoton(ancestorPDG) || MC::isElectron(ancestorPDG) || MC::isTau(ancestorPDG))) && numberOfChildren > 1 && NumOfNucFr != 0) return ElMagProc;

  if (numOfParents == 1 && abs(ancestorPDG) == MC::PIPLUS && numberOfChildren > 2 && NumOfNucFr != 0) return ElMagProc;

  // nuclear photo fission
  if (MC::isPhoton(ancestorPDG) && numberOfChildren > 4 && NumOfNucFr != 0) return ElMagProc;

  // unknown process el(pos)->el+pos??
  if (MC::isElectron(ancestorPDG) && numberOfChildren == 2 && NumOfEl == 1 && NumOfPos == 1) return ElMagProc;

  // unknown process el->el+el??
  if (ancestorPDG == MC::ELECTRON && numberOfChildren == 2 && NumOfEl == 2 && NumOfPos == 0) return ElMagProc;

  // unknown process pos->pos+pos??
  if (ancestorPDG == MC::POSITRON && numberOfChildren == 2 && NumOfEl == 0 && NumOfPos == 2) return ElMagProc;

  // unknown process pos/el->pos/el??
  if (MC::isElectron(ancestorPDG) && !MC::isDecayed(ancestor) && ancestorPDG == thePriPart->pdgId() && numberOfChildren == 1 && !samePart) return ElMagProc;

  // pi->pi+e+/e-; mu->mu+e+/e- ;
  // gamma+ atom->gamma(the same) + e (compton scattering)
  if (numberOfChildren == 2 && (NumOfEl == 1 || NumOfPos == 1) && !MC::isElectron(ancestorPDG) && samePart) return ElMagProc;

  if ((ancestorPDG == MC::PI0 && numberOfChildren == 3 && NumOfPhot == 1 && NumOfEl == 1 && NumOfPos == 1) ||
      (ancestorPDG == MC::PI0 && numberOfChildren == 4 && NumOfPhot == 0 && NumOfEl == 2 && NumOfPos == 2))
    return DalitzDec;

  // Quark weak decay
  if (MC::isSMQuark(ancestorPDG) && numOfParents == 1 && numberOfChildren == 3 && NumOfquark == 1 && NumOfElNeut == 1) return QuarkWeakDec;

  if (MC::isMuon(ancestorPDG) && NumOfNucFr != 0) return ElMagProc;

  if (MC::isTop(ancestorPDG)) return top;

  if (MC::isW(ancestorPDG) && ancestorOrigVert && ancestorOrigVert->nIncomingParticles() != 0) {

    const xAOD::TruthVertex* prodVert = ancestorOrigVert;
    const xAOD::TruthParticle* ptrPart{};
    do {
      ptrPart = prodVert->incomingParticle(0); // FIXME just taking the first one
      prodVert = ptrPart->hasProdVtx() ? ptrPart->prodVtx() : nullptr;
    } while (MC::isW(ptrPart) && prodVert);

    if (prodVert && prodVert->nIncomingParticles() == 1) {
      if (abs(ptrPart->pdgId()) == MC::RH_NU_E) return NuREle; // Right-handed NU_E (Pythia-specific)
      if (abs(ptrPart->pdgId()) == MC::RH_NU_MU) return NuRMu; // Right-handed NU_MU (Pythia-specific)
      if (abs(ptrPart->pdgId()) == MC::RH_NU_TAU) return NuRTau; // Right-handed NU_TAU (Pythia-specific)
    }
    return WBoson;
  }
  if (MC::isW(ancestorPDG)) return WBoson;
  if (MC::isZ(ancestorPDG)) return ZBoson;

  // MadGraphPythia ZWW*->lllnulnu
  if (numOfParents == 1 && numberOfChildren > 4 && (MC::isSMQuark(ancestorPDG) || MC::isGluon(ancestorPDG))) {

    const xAOD::TruthParticle* thePartToCheck = thePriPart;
    const xAOD::TruthParticle* theParent = thePriPart->hasProdVtx() ? thePriPart->prodVtx()->incomingParticle(0) : nullptr; // FIXME just taking the first one
    if (theParent && MC::isElectron(theParent) && MC::isDecayed(theParent)) { thePartToCheck = theParent; }

    bool isZboson = false;
    bool isWboson = false;
    bool skipnext = false;

    for (unsigned int ipOut = 0; ipOut + 1 < partOrigVtx->nOutgoingParticles(); ++ipOut) {
      const xAOD::TruthParticle* aChild = partOrigVtx->outgoingParticle(ipOut);
      if (!aChild) continue;
      const xAOD::TruthParticle* theNextChild{};
      for (unsigned int ipOut1 = ipOut + 1; ipOut1 < partOrigVtx->nOutgoingParticles(); ipOut1++) {
        theNextChild = partOrigVtx->outgoingParticle(ipOut1);
        if (theNextChild) break;
      }
      if (!theNextChild) continue;
      if (skipnext) {
        skipnext = false;
        continue;
      }

      if (MC::isElectron(aChild) && MC::isElectron(theNextChild)) {
        // Zboson
        if (thePartToCheck == aChild || thePartToCheck == theNextChild) {
          isZboson = true;
          break;
        }
        skipnext = true;
      } else if (MC::isElectron(aChild) && abs(theNextChild->pdgId()) == MC::NU_E) {
        // WBoson
        if (thePartToCheck == aChild || thePartToCheck == theNextChild) {
          isWboson = true;
          break;
        }
        skipnext = true;
      }
    }
    if (isWboson) return WBoson;
    if (isZboson) return ZBoson;
  }
  if (numOfParents == 2) {
    const int pdg1 = partOrigVtx->incomingParticle(0)->pdgId();
    const int pdg2 = partOrigVtx->incomingParticle(1)->pdgId();
    //--Sherpa Z->ee
    if ((numberOfChildren - NumOfquark - NumOfgluon) == 2 && NumOfEl == 1 && NumOfPos == 1) return ZBoson;

    //--Sherpa W->enu ??
    if ((numberOfChildren - NumOfquark - NumOfgluon) == 2 && (NumOfEl == 1 || NumOfPos == 1) && NumOfElNeut == 1) return WBoson;

    //--Sherpa ZZ,ZW
    if ((numberOfChildren - NumOfquark - NumOfgluon) == 4 &&
        (NumOfEl + NumOfPos + NumOfMuPl + NumOfMuMin + NumOfTau + NumOfElNeut + NumOfMuNeut + NumOfTauNeut == 4) &&
        (MC::isQuark(pdg1)||MC::isGluon(pdg1)) && (MC::isQuark(pdg2)||MC::isGluon(pdg2))) return DiBoson;

    //--Sherpa VVV -- Note, have to allow for prompt photon radiation or these get lost
    if ((numberOfChildren - NumOfquark - NumOfgluon - NumOfPhot) == 6 &&
        (NumOfEl + NumOfPos + NumOfMuPl + NumOfMuMin + NumOfTau + NumOfElNeut + NumOfMuNeut + NumOfTauNeut == 6) &&
        (MC::isQuark(pdg1)||MC::isGluon(pdg1)) && (MC::isQuark(pdg2)||MC::isGluon(pdg2))) return MultiBoson;
  }

  // New Sherpa Z->ee
  if (partOrigVtx == ancestorOrigVert && partOrigVtx) {
    int NumOfEleLoop = 0;
    int NumOfLepLoop = 0;
    int NumOfEleNeuLoop = 0;
    for (const auto *const pout: partOrigVtx->particles_out()) {
      if (!pout) continue;
      for (const auto *const pin: partOrigVtx->particles_in()) {
        if (!pin) continue;
        if (!HepMC::is_same_particle(pout,pin)) continue;
        if (MC::isElectron(pout)) NumOfEleLoop++;
        if (std::abs(pin->pdgId()) == MC::NU_E) NumOfEleNeuLoop++; // FIXME pin vs pout difference in behaviour compared to MCTruthClassifier::defOrigOfMuon
        if (MC::isSMLepton(pout)) NumOfLepLoop++;
      }
    }
    if (NumOfEleLoop == 2 && NumOfEleNeuLoop == 0) return ZBoson;
    if (NumOfEleLoop == 1 && NumOfEleNeuLoop == 1) return WBoson;
    if ((NumOfEleLoop == 4 && NumOfEleNeuLoop == 0) || (NumOfEleLoop == 3 && NumOfEleNeuLoop == 1) ||
        (NumOfEleLoop == 2 && NumOfEleNeuLoop == 2)) return DiBoson;
    if (NumOfLepLoop == 4) return DiBoson;
  }

  //-- McAtNLo

  if (MC::isHiggs(ancestorPDG)) return Higgs;

  if (MC::isMSSMHiggs(ancestorPDG)) return HiggsMSSM; // MSSM Higgs bosons

  if (MC::isHeavyBoson(ancestorPDG)) return HeavyBoson;  // Heavy bosons( Z', Z'', W'+)

  if (MC::isMuon(ancestorPDG)) return Mu;
  if (MC::isTau(ancestorPDG)) {
    const ParticleOrigin tauOrig = defOrigOfTau(xTruthParticleContainer, ancestor, ancestorPDG, info);
    const ParticleType tautype = defTypeOfTau(tauOrig);
    return (tautype == IsoTau)?tauOrig:TauLep;
  }

  if (abs(ancestorPDG) == MC::WBOSON_LRSM) return WBosonLRSM; // Left-right symmetric model WBoson (Pythia-specific)
  if (abs(ancestorPDG) == MC::RH_NU_E) return NuREle; // Right-handed NU_E (Pythia-specific)
  if (abs(ancestorPDG) == MC::RH_NU_MU) return NuRMu; // Right-handed NU_MU (Pythia-specific)
  if (abs(ancestorPDG) == MC::RH_NU_TAU) return NuRTau; // Right-handed NU_TAU (Pythia-specific)
  if (MC::isLeptoQuark(ancestorPDG) || NumOfLQ != 0) return LQ;
  if (MC::isSUSY(ancestorPDG)) return SUSY;
  if (MC::isBSM(ancestorPDG)) return OtherBSM;

  const ParticleType pType = defTypeOfHadron(ancestorPDG);
  if ((pType == BBbarMesonPart || pType == CCbarMesonPart) && ancestorOrigVert && MC::isHardScatteringVertex(ancestorOrigVert)) isPrompt = true;
  return convHadronTypeToOrig(pType, ancestorPDG);
}


ParticleOrigin MCTruthClassifier::defOrigOfMuon(const xAOD::TruthParticleContainer& xTruthParticleContainer,
                                                const xAOD::TruthParticle* thePart,
                                                bool& isPrompt,
                                                MCTruthPartClassifier::Info& info) const
{
  ATH_MSG_DEBUG("Executing DefOrigOfMuon ");

  // Find the first copy of this particle stored in the xAOD::TruthParticleContainer (i.e. the particle prior to any interactions)
  const xAOD::TruthParticle* thePriPart = MC::findMatching(xTruthParticleContainer, thePart);
  if (!thePriPart) return NonDefined;
  if (!MC::isMuon(thePriPart)) return NonDefined;

  const xAOD::TruthVertex* partOrigVtx = thePriPart->hasProdVtx() ? thePriPart->prodVtx() : nullptr;

  //-- to define muon  outcome status
  info.particleOutCome = defOutComeOfMuon(thePriPart);

  if (!partOrigVtx) return NonDefined;

  int numOfParents = partOrigVtx->nIncomingParticles();
  if (numOfParents > 1) ATH_MSG_DEBUG("DefOrigOfMuon:: muon has more than one parent.");

  const xAOD::TruthParticle* ancestor = MC::findMother(thePriPart);
  info.setMotherProperties(ancestor);
  if (!ancestor) { return NonDefined; } // ancestor is not a nullptr beyond this point

  int ancestorPDG = ancestor->pdgId();
  // "method 1" for finding Sherpa loops from defOrigOfElectron not used here. Why?
  const xAOD::TruthVertex* ancestorProdVtx = ancestor->hasProdVtx() ? ancestor->prodVtx() : nullptr;

  if ((MC::isTau(ancestorPDG)|| MC::isW(ancestorPDG)) && ancestorProdVtx) {
    int pPDG(0);
    const xAOD::TruthParticle* AncestorParent{};
    do {
      pPDG = 0;
      AncestorParent = MC::findMother(ancestor);
      // Start of method 2 of protecting against loops
      // to prevent Sherpa loop
      if (ancestor == AncestorParent) { break; }
      if (TruthLoopDetectionMethod2(ancestor,AncestorParent)) {
        AncestorParent = ancestor;
        break;
      }
      // End of method 2 of protecting against loops

      if (AncestorParent) {
        pPDG = AncestorParent->pdgId();// Only set pPDG in the case that we aren't in a loop.
        if (MC::isMuon(pPDG) || MC::isTau(pPDG) || MC::isW(pPDG)) { // FIXME should this be (MC::isTau(pPDG) || MC::isW(pPDG)) ???
          // There will be another iteration so set ancestor to AncestorParent
          ancestor = AncestorParent; // AncestorParent is not nullptr here
        }
      }
    } while ((MC::isMuon(pPDG) || MC::isTau(pPDG) || MC::isW(pPDG))); // FIXME should this be (MC::isTau(pPDG) || MC::isW(pPDG)) ???

    if (MC::isTau(pPDG) || MC::isW(pPDG) || MC::isZ(pPDG) || MC::isHiggs(pPDG) ||
        MC::isMSSMHiggs(pPDG) || MC::isHeavyBoson(pPDG) || MC::isTop(pPDG) || // MSSM Higgs bosons, Heavy bosons( Z', Z'', W'+)
        abs(pPDG) == MC::WBOSON_LRSM || MC::isNeutrinoRH(pPDG) ||  // Left-right symmetric model WBoson || Right-handed neutrino (Pythia-specific)
        MC::isSUSY(pPDG)) {
      info.setMotherProperties(ancestor); // FIXME difference in behaviour compared to MCTruthClassifier::defOrigOfElectron/Neutrino
    }
  }

  info.setMotherProperties(ancestor);
  ancestorPDG = ancestor->pdgId();
  ancestorProdVtx = ancestor->hasProdVtx() ? ancestor->prodVtx() : nullptr;
  partOrigVtx = ancestor->decayVtx();
  numOfParents = partOrigVtx->nIncomingParticles();
  const int numberOfChildren = partOrigVtx->nOutgoingParticles();

  // Determine decay products
  auto DP = DecayProducts(partOrigVtx);
  const int NumOfPhot = DP.pd(MC::PHOTON);
  const int NumOfEl = DP.pd(MC::ELECTRON);
  const int NumOfPos = DP.pd(MC::POSITRON);
  const int NumOfElNeut = DP.apd(MC::NU_E);
  const int NumOfMuNeut = DP.apd(MC::NU_MU);
  const int NumOfLQ = DP.apd(MC::LEPTOQUARK);
  const int NumOfquark = DP.apd({MC::DQUARK,MC::UQUARK,MC::SQUARK,MC::CQUARK,MC::BQUARK,MC::TQUARK});
  const int NumOfgluon = DP.apd(MC::GLUON);
  const int NumOfMuPl = DP.pd(-MC::MUON);
  const int NumOfMuMin = DP.pd(MC::MUON);
  const int NumOfTau = DP.apd(MC::TAU);
  const int NumOfTauNeut = DP.apd(MC::NU_TAU);
  // End of section determining decay products

  if (std::abs(ancestorPDG) == MC::PIPLUS && numberOfChildren == 2 && NumOfMuNeut == 1) return PionDecay;
  if (std::abs(ancestorPDG) == MC::KPLUS && numberOfChildren == 2 && NumOfMuNeut == 1) return KaonDecay;
  if (MC::isTau(ancestorPDG)) {
    const ParticleOrigin tauOrig = defOrigOfTau(xTruthParticleContainer, ancestor, ancestorPDG, info);
    const ParticleType tautype = defTypeOfTau(tauOrig);
    return  (tautype == IsoTau)?tauOrig:TauLep;
  }

  if (MC::isTop(ancestorPDG)) return top;
  // Quark weak decay
  if (MC::isSMQuark(ancestorPDG) && numOfParents == 1 && numberOfChildren == 3 && NumOfquark == 1 && NumOfMuNeut == 1) return QuarkWeakDec;

  if (MC::isW(ancestorPDG) && ancestorProdVtx && ancestorProdVtx->nIncomingParticles() != 0) {
    const xAOD::TruthVertex* prodVert = ancestorProdVtx;
    const xAOD::TruthParticle* itrP;
    do {
      itrP = prodVert->incomingParticle(0); // FIXME just taking the first one
      prodVert = itrP->hasProdVtx() ? itrP->prodVtx() : nullptr;
    } while (MC::isW(itrP) && prodVert);

    if (prodVert && prodVert->nIncomingParticles() == 1) {
      if (abs(itrP->pdgId()) == MC::RH_NU_E) return NuREle; // Right-handed NU_E (Pythia-specific)
      if (abs(itrP->pdgId()) == MC::RH_NU_MU) return NuRMu; // Right-handed NU_MU (Pythia-specific)
      if (abs(itrP->pdgId()) == MC::RH_NU_TAU) return NuRTau; // Right-handed NU_TAU (Pythia-specific)
    }
    return WBoson;
  }
  if (MC::isW(ancestorPDG)) return WBoson;
  if (MC::isZ(ancestorPDG)) return ZBoson;
  if (MC::isPhoton(ancestorPDG) && numberOfChildren == 2 && NumOfMuMin == 1 && NumOfMuPl == 1) return PhotonConv;
  //-- Exotics

  // MadGraphPythia ZWW*->lllnulnu
  if (numOfParents == 1 && numberOfChildren > 4 && (MC::isSMQuark(ancestorPDG) || MC::isGluon(ancestorPDG))) {
    bool isZboson = false;
    bool isWboson = false;
    bool skipnext = false;
    for (unsigned int ipOut = 0; ipOut + 1 < partOrigVtx->nOutgoingParticles(); ipOut++) {
      if (skipnext) {
        skipnext = false;
        continue;
      }
      const xAOD::TruthParticle* aChild = partOrigVtx->outgoingParticle(ipOut);
      if (!aChild) continue;
      const xAOD::TruthParticle* theNextChild{};
      for (unsigned int ipOut1 = ipOut + 1; ipOut1 < partOrigVtx->nOutgoingParticles(); ipOut1++) {
        theNextChild = partOrigVtx->outgoingParticle(ipOut1);
        if (theNextChild) break;
      }
      if (!theNextChild) continue;
      if (MC::isMuon(aChild) && MC::isMuon(theNextChild)) {
        // Zboson
        if (thePriPart == aChild || thePriPart == theNextChild) {
          isZboson = true;
          break;
        }
        skipnext = true;
      } else if (MC::isMuon(aChild) && abs(theNextChild->pdgId()) == MC::NU_MU) {
        // WBoson
        if (thePriPart == aChild || thePriPart == theNextChild) {
          isWboson = true;
          break;
        }
        skipnext = true;
      }
    }
    if (isWboson) return WBoson;
    if (isZboson) return ZBoson;
  }
  if (numOfParents == 2 ) {
    const int pdg1 = partOrigVtx->incomingParticle(0)->pdgId();
    const int pdg2 = partOrigVtx->incomingParticle(1)->pdgId();
    //--Sherpa Z->mumu
    if ((numberOfChildren - NumOfquark - NumOfgluon) == 2 && NumOfMuPl == 1 && NumOfMuMin == 1) return ZBoson;

    //--Sherpa W->munu ??
    // if(numOfParents==2&&(numberOfChildren-NumOfquark-NumOfgluon)==2&&(NumOfEl==1||NumOfPos==1)&&NumOfElNeut==1) return WBoson;
    if ((numberOfChildren - NumOfquark - NumOfgluon) == 2 && (NumOfMuPl == 1 || NumOfMuMin == 1) && NumOfMuNeut == 1) return WBoson;

    //--Sherpa ZZ,ZW
    if ((numberOfChildren - NumOfquark - NumOfgluon) == 4 &&
        (NumOfEl + NumOfPos + NumOfMuPl + NumOfMuMin + NumOfTau + NumOfElNeut + NumOfMuNeut + NumOfTauNeut == 4) &&
        (MC::isQuark(pdg1)||MC::isGluon(pdg1)) && (MC::isQuark(pdg2)||MC::isGluon(pdg2))) return DiBoson;

    //--Sherpa VVV -- Note, have to allow for prompt photon radiation or these get lost
    if ((numberOfChildren - NumOfquark - NumOfgluon - NumOfPhot) == 6 &&
        (NumOfEl + NumOfPos + NumOfMuPl + NumOfMuMin + NumOfTau + NumOfElNeut + NumOfMuNeut + NumOfTauNeut == 6) &&
        (MC::isQuark(pdg1)||MC::isGluon(pdg1)) && (MC::isQuark(pdg2)||MC::isGluon(pdg2))) return MultiBoson;
  }

  //--New Sherpa Z->mumu
  if (partOrigVtx == ancestorProdVtx) {
    int NumOfMuLoop = 0;
    int NumOfMuNeuLoop = 0;
    int NumOfLepLoop = 0;
    for (const auto & pout: partOrigVtx->particles_out()) {
      for (const auto & pin: partOrigVtx->particles_in()) {
        if (!pout) continue;
        if (!pin) continue;
        if (HepMC::is_same_particle(pout,pin)) {
          if (MC::isMuon(pout)) NumOfMuLoop++;
          if (std::abs(pout->pdg_id()) == MC::NU_MU) NumOfMuNeuLoop++; // FIXME pin vs pout difference in behaviour compared to MCTruthClassifier::defOrigOfElectron
          if (MC::isSMLepton(pout)) NumOfLepLoop++;
        }
      }
    }
    if (NumOfMuLoop == 2 && NumOfMuNeuLoop == 0) return ZBoson;
    if (NumOfMuLoop == 1 && NumOfMuNeuLoop == 1) return WBoson;
    if ((NumOfMuLoop == 4 && NumOfMuNeuLoop == 0) || (NumOfMuLoop == 3 && NumOfMuNeuLoop == 1) ||
        (NumOfMuLoop == 2 && NumOfMuNeuLoop == 2)) return DiBoson;
    if (NumOfLepLoop == 4) return DiBoson;
  }

  //-- McAtNLo

  if (MC::isHiggs(ancestorPDG)) return Higgs;

  if (MC::isMSSMHiggs(ancestorPDG)) return HiggsMSSM; // MSSM Higgs bosons

  if (MC::isHeavyBoson(ancestorPDG))  return HeavyBoson;  // Heavy bosons( Z', Z'', W'+)

  if (abs(ancestorPDG) == MC::WBOSON_LRSM) return WBosonLRSM; // Left-right symmetric model WBoson (Pythia-specific)
  if (abs(ancestorPDG) == MC::RH_NU_E) return NuREle; // Right-handed NU_E (Pythia-specific)
  if (abs(ancestorPDG) == MC::RH_NU_MU) return NuRMu; // Right-handed NU_MU (Pythia-specific)
  if (abs(ancestorPDG) == MC::RH_NU_TAU) return NuRTau; // Right-handed NU_TAU (Pythia-specific)
  if (MC::isLeptoQuark(ancestorPDG) || NumOfLQ != 0) return LQ;
  if (MC::isSUSY(ancestorPDG)) return SUSY;
  if (MC::isBSM(ancestorPDG)) return OtherBSM;

  const ParticleType pType = defTypeOfHadron(ancestorPDG);
  if ((pType == BBbarMesonPart || pType == CCbarMesonPart) && ancestorProdVtx && MC::isHardScatteringVertex(ancestorProdVtx)) isPrompt = true;

  return convHadronTypeToOrig(pType, ancestorPDG);
}

ParticleOrigin MCTruthClassifier::defOrigOfTau(const xAOD::TruthParticleContainer& xTruthParticleContainer,
                                               const xAOD::TruthParticle* thePart,
                                               int ancestorPDGin,
                                               MCTruthPartClassifier::Info& info) const
{
  ATH_MSG_DEBUG("Executing DefOrigOfTau ");

  // Find the first copy of this particle stored in the xAOD::TruthParticleContainer (i.e. the particle prior to any interactions)
  const xAOD::TruthParticle* thePriPart = MC::findMatching(xTruthParticleContainer, thePart);

  if (!thePriPart) return NonDefined;
  if (!MC::isTau(thePriPart)) return NonDefined;

  const xAOD::TruthVertex* partOrigVtx = thePriPart->hasProdVtx() ? thePriPart->prodVtx() : nullptr;

  //-- to define tau  outcome status
  if (MC::isPhysical(thePriPart)) info.particleOutCome = defOutComeOfTau(thePriPart); // FIXME why do we need the additional check on MC::isPhysical here c.f. defOrigOfElectron and defOrigOfMuon?

  if (!partOrigVtx) return NonDefined;

  int numOfParents = partOrigVtx->nIncomingParticles();
  if (numOfParents > 1) ATH_MSG_DEBUG("DefOrigOfTau:: tau has more than one parent.");

  const xAOD::TruthParticle* ancestor = MC::findMother(thePriPart);
  info.setMotherProperties(ancestor);
  if (!ancestor) { return NonDefined; } // ancestor is not a nullptr beyond this point

  // "method 1" for finding Sherpa loops from defOrigOfElectron not used here. Why?
  const xAOD::TruthVertex* ancestorProdVtx = ancestor->hasProdVtx() ? ancestor->prodVtx() : nullptr;

  // Difference from defOrigOfElectron and defOrigOfMuon - no loop through ancestor particles
  const xAOD::TruthParticle* AncestorParent{};

  if (MC::isW(ancestorPDGin) && ancestorProdVtx) { // FIXME ancestorPDGin here could in principle be inconsistent with ancestorProdVtx
    AncestorParent = MC::findMother(ancestor);
    if (AncestorParent) {//AncestorParent checked here...
      if (MC::isTop(AncestorParent->pdgId())) {
        ancestor = AncestorParent; //...so ancestor cannot be nullptr
      }
    }
  }

  int ancestorPDG = ancestor->pdgId();
  info.setMotherProperties(ancestor);
  ancestorProdVtx = ancestor->hasProdVtx() ? ancestor->prodVtx() : nullptr;
  partOrigVtx = ancestor->decayVtx();
  if (!partOrigVtx) return NonDefined; // FIXME not sure this could ever be true?

  numOfParents = partOrigVtx->nIncomingParticles();

  // Determine decay products
  auto DP = DecayProducts(partOrigVtx);
  const int numberOfChildren = DP.size();
  const int NumOfPhot = DP.pd(MC::PHOTON);
  const int NumOfEl = DP.pd(MC::ELECTRON);
  const int NumOfPos = DP.pd(MC::POSITRON);
  const int NumOfElNeut = DP.apd(MC::NU_E);
  const int NumOfMuNeut = DP.apd(MC::NU_MU);
  /* const int NumOfLQ = DP.apd(MC::LEPTOQUARK); */ // FIXME Leptoquarks not an option?
  const int NumOfquark = DP.apd({MC::DQUARK,MC::UQUARK,MC::SQUARK,MC::CQUARK,MC::BQUARK,MC::TQUARK});
  const int NumOfgluon = DP.apd(MC::GLUON);
  const int NumOfMuPl = DP.pd(-MC::MUON);
  const int NumOfMuMin = DP.pd(MC::MUON);
  const int NumOfTau = DP.apd(MC::TAU);
  const int NumOfTauNeut = DP.apd(MC::NU_TAU);
  // End of section determining decay products

  if (MC::isTop(ancestorPDG)) return top;
  if (MC::isW(ancestorPDG) && ancestorProdVtx && ancestorProdVtx->nIncomingParticles() != 0) {
    const xAOD::TruthVertex* prodVert = ancestorProdVtx;
    const xAOD::TruthParticle* itrP;
    do {
      itrP = prodVert->incomingParticle(0); // FIXME just taking the first one
      prodVert = itrP->hasProdVtx() ? itrP->prodVtx() : nullptr;
    } while (MC::isW(itrP) && prodVert);

    if (prodVert && prodVert->nIncomingParticles() == 1 ) {
      if (abs(itrP->pdgId()) == MC::RH_NU_E) return NuREle; // Right-handed NU_E (Pythia-specific)
      if (abs(itrP->pdgId()) == MC::RH_NU_MU) return NuRMu; // Right-handed NU_MU (Pythia-specific)
      if (abs(itrP->pdgId()) == MC::RH_NU_TAU) return NuRTau; // Right-handed NU_TAU (Pythia-specific)
    }
    return WBoson;
  }
  if (MC::isW(ancestorPDG)) { return WBoson;}
  if (MC::isZ(ancestorPDG)) { return ZBoson;}
  if (numOfParents == 1 && numberOfChildren > 4 && (MC::isSMQuark(ancestorPDG) || MC::isGluon(ancestorPDG))) {
    bool isZboson = false;
    bool isWboson = false;
    bool skipnext = false;
    for (unsigned int ipOut = 0; ipOut + 1 < partOrigVtx->nOutgoingParticles(); ipOut++) {
      if (skipnext) {
        skipnext = false;
        continue;
      }
      const xAOD::TruthParticle* aChild = partOrigVtx->outgoingParticle(ipOut);
      if (!aChild) continue;
      const xAOD::TruthParticle* theNextChild{};
      for (unsigned int ipOut1 = ipOut + 1; ipOut1 < partOrigVtx->nOutgoingParticles(); ipOut1++) {
        theNextChild = partOrigVtx->outgoingParticle(ipOut1);
        if (theNextChild) break;
      }
      if (!theNextChild) {
        continue;
      }
      if (MC::isTau(aChild) && MC::isTau(theNextChild)) {
        // Zboson
        if (thePriPart == aChild || thePriPart == theNextChild) {
          isZboson = true;
          break;
        }
        skipnext = true;
      } else if (MC::isTau(aChild) && abs(theNextChild->pdgId()) == MC::NU_TAU) {
        // WBoson
        if (thePriPart == aChild || thePriPart == theNextChild) {
          isWboson = true;
          break;
        }
        skipnext = true;
      }
    }
    if (isWboson) return WBoson;
    if (isZboson) return ZBoson;
  }
  if (numOfParents == 2 ) {
    const int pdg1 = partOrigVtx->incomingParticle(0)->pdgId();
    const int pdg2 = partOrigVtx->incomingParticle(1)->pdgId();
    //--Sherpa Z->tautau
    if ((numberOfChildren - NumOfquark - NumOfgluon) == 2 && NumOfTau == 2  && (MC::isQuark(pdg1)||MC::isGluon(pdg1)) && (MC::isQuark(pdg2)||MC::isGluon(pdg2))) return ZBoson;

    //--Sherpa W->taunu  new
    if ((numberOfChildren - NumOfquark - NumOfgluon) == 2 && NumOfTau == 1 && NumOfTauNeut == 1) return WBoson;

    //--Sherpa ZZ,ZW
    if ((numberOfChildren - NumOfquark - NumOfgluon) == 4 &&
        (NumOfEl + NumOfPos + NumOfMuPl + NumOfMuMin + NumOfTau + NumOfElNeut + NumOfMuNeut + NumOfTauNeut == 4) &&
        (MC::isQuark(pdg1)||MC::isGluon(pdg1)) && (MC::isQuark(pdg2)||MC::isGluon(pdg2))) return DiBoson;

    //--Sherpa VVV -- Note, have to allow for prompt photon radiation or these get lost
    if ((numberOfChildren - NumOfquark - NumOfgluon - NumOfPhot) == 6 &&
        (NumOfEl + NumOfPos + NumOfMuPl + NumOfMuMin + NumOfTau + NumOfElNeut + NumOfMuNeut + NumOfTauNeut == 6) &&
        (MC::isQuark(pdg1)||MC::isGluon(pdg1)) && (MC::isQuark(pdg2)||MC::isGluon(pdg2))) return MultiBoson;
  }

  // New Sherpa Z->tautau
  if (partOrigVtx == ancestorProdVtx) {
    int NumOfTauLoop = 0;
    int NumOfTauNeuLoop = 0;
    int NumOfLepLoop = 0;
    for ( const auto *const pout: partOrigVtx->particles_out()) {
      if (!pout) continue;
      for (const auto *const pin: partOrigVtx->particles_in()) {
        if (!pin) continue;
        if (!HepMC::is_same_particle(pout,pin)) continue;
        if (MC::isTau(pout)) NumOfTauLoop++;
        if (std::abs(pout->pdgId()) == MC::NU_TAU) NumOfTauNeuLoop++; // FIXME pin vs pout difference in behaviour compared to MCTruthClassifier::defOrigOfElectron
        if (MC::isSMLepton(pout)) NumOfLepLoop++;
      }
    }
    if (NumOfTauLoop == 2 && NumOfTauNeuLoop == 0) return ZBoson;
    if (NumOfTauLoop == 1 && NumOfTauNeuLoop == 1) return WBoson;
    if ((NumOfTauLoop == 4 && NumOfTauNeuLoop == 0) || (NumOfTauLoop == 3 && NumOfTauNeuLoop == 1) || (NumOfTauLoop == 2 && NumOfTauNeuLoop == 2)) return DiBoson;
    if (NumOfLepLoop == 4) return DiBoson;
  }

  //-- McAtNLo

  if (MC::isHiggs(ancestorPDG)) return Higgs;
  if (MC::isMSSMHiggs(ancestorPDG)) return HiggsMSSM; // MSSM Higgs bosons
  if (MC::isHeavyBoson(ancestorPDG)) return HeavyBoson; // Heavy bosons( Z', Z'', W'+)
  if (abs(ancestorPDG) == MC::WBOSON_LRSM) return WBosonLRSM; // Left-right symmetric model WBoson (Pythia-specific)
  if (abs(ancestorPDG) == MC::RH_NU_TAU) return NuRTau; // Right-handed NU_TAU (Pythia-specific)
  if (MC::isSUSY(ancestorPDG)) return SUSY;
  if (MC::isBSM(ancestorPDG)) return OtherBSM;
  if (abs(ancestorPDG) == MC::JPSI) return JPsi;

  const ParticleType pType = defTypeOfHadron(ancestorPDG);
  return convHadronTypeToOrig(pType, ancestorPDG);
}

ParticleOrigin MCTruthClassifier::defOrigOfPhoton(const xAOD::TruthParticleContainer& xTruthParticleContainer,
                                                  const xAOD::TruthParticle* thePart,
                                                  bool& isPrompt,
                                                  MCTruthPartClassifier::Info& info) const
{
  if (!thePart) return NonDefined; // FIXME Why is this extra protection needed for this function and not the others?
  ATH_MSG_DEBUG("Executing DefOrigOfPhoton ");

  info.resetMotherProperties();
  info.photonMother = nullptr;

  // Find the first copy of this particle stored in the xAOD::TruthParticleContainer (i.e. the particle prior to any interactions)
  const xAOD::TruthParticle* thePriPart = MC::findMatching(xTruthParticleContainer, thePart);
  if (!thePriPart) return NonDefined;
  if (!MC::isPhoton(thePriPart)) return NonDefined;

  const xAOD::TruthVertex* partOrigVtx = thePriPart->hasProdVtx() ? thePriPart->prodVtx() : nullptr;

  //-- to define photon outcome status
  info.particleOutCome = defOutComeOfPhoton(thePriPart);

  if (!partOrigVtx) return NonDefined;

  int numOfParents = partOrigVtx->nIncomingParticles();
  if (partOrigVtx->nIncomingParticles() > 1) ATH_MSG_DEBUG("DefOrigOfPhoton:: photon has more than one parent.");

  const xAOD::TruthParticle* ancestor = MC::findMother(thePriPart);
  info.setMotherProperties(ancestor);
  if (!ancestor) { return NonDefined; } // ancestor is not a nullptr beyond this point

  int ancestorPDG = ancestor->pdgId();
  // "method 1" for finding Sherpa loops from defOrigOfElectron not used here. Why?
  const xAOD::TruthVertex* ancestorProdVtx = ancestor->hasProdVtx() ? ancestor->prodVtx() : nullptr;

  partOrigVtx = ancestor->decayVtx(); // FIXME how often does this line actually change the pointer???
  numOfParents = partOrigVtx->nIncomingParticles();

  const int numberOfChildren = partOrigVtx->nOutgoingParticles();

  // Determine decay products
  auto DP = DecayProducts(partOrigVtx);
  const int NumOfEl = DP.pd(MC::ELECTRON);
  const int NumOfPos = DP.pd(MC::POSITRON);
  const int NumOfMu = DP.apd(MC::MUON);
  const int NumOfTau = DP.apd(MC::TAU);
  const int NumOfLQ = DP.apd(MC::LEPTOQUARK);
  const long NumOfLep = NumOfEl + NumOfPos + NumOfMu + NumOfTau; // FIXME does this really need to be a long???
  const long NumOfNeut = DP.apd({MC::NU_E,MC::NU_MU,MC::NU_TAU}); // FIXME does this really need to be a long???
  const int NumOfPht = DP.pd(MC::PHOTON);

  int childPDG(0);
  long NumOfPartons(0); // FIXME does this really need to be a long???
  int NumOfNucFr(0);
  const xAOD::TruthParticle* child{};
  const bool possibleNuclearFragment = (numOfParents == 1 && (MC::isPhoton(ancestorPDG) || MC::isElectron(ancestorPDG) || abs(ancestorPDG) == MC::PIPLUS));
  for (const auto& pout: partOrigVtx->particles_out()) {
    if (!pout) continue;
    childPDG = pout->pdg_id();
    if (possibleNuclearFragment &&
        (MC::isNucleus(childPDG) || childPDG == 0 || childPDG == MC::PROTON || childPDG == MC::NEUTRON)) { // FIXME Do we really expect particles with PDG_ID = 0 in the truth record?
      NumOfNucFr++;
    }
    if (abs(childPDG) < MC::ELECTRON ||
        (abs(childPDG) > MC::NU_TAU && abs(childPDG) < 43 && !MC::isPhoton(childPDG))) {
      // FIXME Too loose? This definition picks up 4th generation quarks and leptons as well as all gauge bosons and leptoquarks.
      // Suggest MC::isSMQuark(childPDG) || (MC::isBoson(childPDG) && !MC::isPhoton(childPDG))
      // or maybe even MC::isSMQuark(childPDG) || MC::isGluon(childPDG)
      // AKA const long NumOfPartons = DP.apd({MC::DQUARK,MC::UQUARK,MC::SQUARK,MC::CQUARK,MC::BQUARK,MC::TQUARK,MC::GLUON});
      NumOfPartons++;
    }
    if (childPDG == ancestorPDG) {
      child = pout;
    }
  }
  // End of section determining decay products

  bool foundISR = false;
  bool foundFSR = false;
  if (numOfParents == 1 && numberOfChildren == 2 &&  child && HepMC::is_same_generator_particle(child, ancestor)) return BremPhot;
  if (numOfParents == 1 && numberOfChildren == 2 && MC::isElectron(ancestorPDG) && NumOfPht == 2) return ElMagProc;

  // decay of W,Z and Higgs to lepton with FSR generated by Pythia
  if (numOfParents == 1 && numberOfChildren == 2 && (MC::isElectron(ancestorPDG) || MC::isMuon(ancestorPDG) || MC::isTau(ancestorPDG)) &&
      !(child && HepMC::is_same_generator_particle(child, ancestor)) && ancestorProdVtx &&
      ancestorProdVtx->nIncomingParticles() == 1) {
    int itr = 0;
    int PartPDG = 0;
    const xAOD::TruthVertex* prodVert = ancestorProdVtx;
    const xAOD::TruthVertex* Vert{};
    do {
      Vert = prodVert;
      for (const auto & pin: Vert->particles_in()) {
        if (!pin) continue;
        PartPDG = abs(pin->pdgId());
        prodVert = pin->prodVtx();
        if (MC::isZ(PartPDG) || MC::isW(PartPDG) || MC::isHiggs(PartPDG)) foundFSR = true;
      }
      itr++;
      if (itr > 100) { // FIXME Improve loop detection here?
        ATH_MSG_WARNING("DefOrigOfPhoton:: infinite while");
        break;
      }
    } while (prodVert && abs(ancestorPDG) == PartPDG);

    if (foundFSR) return FSRPhot;
  }

  // Nucl reaction
  // gamma+Nuclear=>gamma+Nucl.Fr+Nuclons+pions
  // e+Nuclear=>e+gamma+Nucl.Fr+Nuclons+pions
  // pi+Nuclear=>gamma+Nucl.Fr+Nuclons+pions

  if ((numOfParents == 1 && (MC::isPhoton(ancestorPDG) || MC::isElectron(ancestorPDG)) && numberOfChildren > 2 && NumOfNucFr != 0) ||
      (numOfParents == 1 && abs(ancestorPDG) == MC::PIPLUS && numberOfChildren > 10 && NumOfNucFr != 0) ||
      (numOfParents == 1 && MC::isPhoton(ancestorPDG) && numberOfChildren > 10 && MC::isStable(ancestor)) ||
      (numOfParents == 1 && MC::isNucleus(ancestorPDG) && std::abs(ancestorPDG) != MC::PROTON)) // FIXME vetoing protons here to preserve previous behaviour
    return NucReact;

  if (MC::isMuon(ancestorPDG) && NumOfMu == 0) return Mu;
  if (MC::isTau(ancestorPDG) && NumOfTau == 0) return TauLep;

  if (numOfParents == 1 && ancestor->status() == 3) return (foundISR)? ISRPhot:UndrPhot; // FIXME foundISR is always false at this point

  //-- to find initial and final state raiation and underline photons
  //-- SUSY
  if (numOfParents == 1 && (MC::isSMQuark(ancestorPDG) || MC::isGluon(ancestorPDG)) &&
      (numberOfChildren != NumOfPht + NumOfPartons || !MC::Pythia8::isConditionA(ancestor))) {
    for (const auto& pout: partOrigVtx->particles_out()) {
      if (!pout) continue;
      if (ancestorPDG != pout->pdgId()) continue;
      const xAOD::TruthVertex* Vrtx = pout->decayVtx();
      if (!Vrtx) continue;
      if (Vrtx->nOutgoingParticles() != 1 && Vrtx->nIncomingParticles() == 1) continue;
      if (!Vrtx->outgoingParticle(0)) continue;
      if (Vrtx->outgoingParticle(0)->pdgId() == 91) foundISR = true; // Herwig "cluster"
    }
    return (foundISR)?ISRPhot:UndrPhot;
  }

  //-- to find final  state radiation
  //-- Exotics

  // FSR  from Photos
  //-- Exotics- CompHep
  if (numOfParents == 2 && ((MC::isElectron(ancestorPDG) && NumOfEl == 1 && NumOfPos == 1) || (MC::isMuon(ancestorPDG) && NumOfMu == 2) || (MC::isTau(ancestorPDG) && NumOfTau == 2))) {
    if (abs(partOrigVtx->incomingParticle(0)->pdgId()) == abs(partOrigVtx->incomingParticle(1)->pdgId())) return FSRPhot;
  }

  if (numOfParents == 2 && NumOfLep == 1 && NumOfNeut == 1 && (MC::isElectron(ancestorPDG) || abs(ancestorPDG) == MC::NU_E)) return FSRPhot;

  //-- Exotics - CompHep
  if (MC::isElectron(ancestorPDG) && numOfParents == 1 && numberOfChildren == 2 && (NumOfEl == 1 || NumOfPos == 1) && NumOfPht == 1 &&
      !( child && HepMC::is_same_generator_particle(child, ancestor)) && !HepMC::is_simulation_particle(child) && !HepMC::is_simulation_particle(ancestor))
    return FSRPhot;

  // FSR  from Photos
  if (MC::isZ(ancestorPDG) && ((NumOfEl + NumOfPos == 2 || NumOfEl + NumOfPos == 4) || (NumOfMu == 2 || NumOfMu == 4) || (NumOfTau == 2 || NumOfTau == 4)) && NumOfPht > 0) return FSRPhot;

  if (NumOfPht > 0 && (abs(ancestorPDG) == MC::WBOSON_LRSM || MC::isNeutrinoRH(ancestorPDG))) return FSRPhot; // Left-right symmetric model WBoson || Right-handed neutrinos (Pythia-specific)

  if (numOfParents == 2 && NumOfLQ == 1) return FSRPhot;

  //--- other process

  if (MC::isZ(ancestorPDG)) return ZBoson;
  if (MC::isW(ancestorPDG)) {

    if (NumOfLep == 1 && NumOfNeut == 1 && numberOfChildren == NumOfLep + NumOfNeut + NumOfPht) return FSRPhot;

    if (ancestorProdVtx && ancestorProdVtx->nIncomingParticles() != 0) {
      const xAOD::TruthVertex* prodVert = ancestorProdVtx;
      const xAOD::TruthParticle* itrP;
      do {
        itrP = prodVert->incomingParticle(0); // FIXME just taking the first one
        prodVert = itrP->hasProdVtx() ? itrP->prodVtx() : nullptr;
      } while (MC::isW(itrP) && prodVert);

      if (prodVert && prodVert->nIncomingParticles() == 1 ) {
        if ( MC::isTau(itrP)) return TauLep;
        if ( MC::isMuon(itrP)) return Mu;
        if ( abs(itrP->pdgId()) == MC::RH_NU_E) return NuREle; // Right-handed NU_E (Pythia-specific)
        if ( abs(itrP->pdgId()) == MC::RH_NU_MU) return NuRMu; // Right-handed NU_MU (Pythia-specific)
        if ( abs(itrP->pdgId()) == MC::RH_NU_TAU) return NuRTau; // Right-handed NU_TAU (Pythia-specific)
      }
    } else
      return WBoson;
  }

  // MadGraphPythia ZWW*->lllnulnu
  if (numOfParents == 1 && numberOfChildren > 4 && (MC::isSMQuark(ancestorPDG) || MC::isGluon(ancestorPDG))) {
    bool isZboson = false;
    bool isWboson = false;
    bool skipnext = false;
    for (unsigned int ipOut = 0; ipOut + 1 < partOrigVtx->nOutgoingParticles(); ipOut++) {
      if (skipnext) {
        skipnext = false;
        continue;
      }
      const xAOD::TruthParticle* aChild = partOrigVtx->outgoingParticle(ipOut);
      if (!aChild) continue;
      const xAOD::TruthParticle* theNextChild{};
      for (unsigned int ipOut1 = ipOut + 1; ipOut1 < partOrigVtx->nOutgoingParticles(); ipOut1++) {
        theNextChild = partOrigVtx->outgoingParticle(ipOut1);
        if (theNextChild) break;
      }
      if (!theNextChild) continue;
      if (MC::isTau(aChild) && MC::isTau(theNextChild)) {
        // Zboson
        if (thePriPart == aChild || thePriPart == theNextChild) {
          isZboson = true;
          break;
        }
        skipnext = true;
      } else if (MC::isTau(aChild) && abs(theNextChild->pdgId()) == MC::NU_TAU) {
        // WBoson
        if (thePriPart == aChild || thePriPart == theNextChild) {
          isWboson = true;
          break;
        }
        skipnext = true;
      }
    }
    if (isWboson) return WBoson;
    if (isZboson) return ZBoson;
  }

  //--Sherpa ZZ,ZW+FSR
  if (numOfParents == 4 && (numberOfChildren - NumOfPht) == 4 && (NumOfLep + NumOfNeut == 4)) {
    if (MC::isSMLepton(partOrigVtx->incomingParticle(0))&&MC::isSMLepton(partOrigVtx->incomingParticle(1))
        && MC::isSMLepton(partOrigVtx->incomingParticle(2))&&MC::isSMLepton(partOrigVtx->incomingParticle(3)))
      return FSRPhot;
  }

  //--New Sherpa single photon
  if (partOrigVtx == ancestorProdVtx && partOrigVtx) {
    int NumOfPhtLoop = 0;
    for (const auto *const pout: partOrigVtx->particles_out()) {
      if (!pout) continue;
      for (const auto *const pin: partOrigVtx->particles_in()) {
        if (!pin) continue;
        if (HepMC::is_same_particle(pout,pin) && MC::isPhoton(pout)) NumOfPhtLoop++;
        if (NumOfPhtLoop == 1) return SinglePhot;
      }
    }
  }

  if (MC::isHiggs(ancestorPDG)) return Higgs;
  if (abs(ancestorPDG) == MC::PI0) return PiZero;
  if (MC::isMSSMHiggs(ancestorPDG)) return HiggsMSSM; // MSSM Higgs bosons
  if (MC::isHeavyBoson(ancestorPDG) || std::abs(ancestorPDG) == 5100039 ) return HeavyBoson; // Heavy Bosons (Z' Z'' W'+) + KK excited graviton

  if (MC::isSUSY(ancestorPDG)) return SUSY;
  if (MC::isBSM(ancestorPDG)) return OtherBSM;

  // Pythia8 gamma+jet samples
  if (MC::Pythia8::isConditionA(ancestor) && MC::isStable(thePriPart) && NumOfPht == 1 && numberOfChildren == (NumOfPht + NumOfPartons))  return PromptPhot;

  const ParticleType pType = defTypeOfHadron(ancestorPDG);
  if ((pType == BBbarMesonPart || pType == CCbarMesonPart) && ancestorProdVtx && MC::isHardScatteringVertex(ancestorProdVtx)) isPrompt = true;
  return convHadronTypeToOrig(pType, ancestorPDG);
}

ParticleOrigin
MCTruthClassifier::defOrigOfNeutrino(const xAOD::TruthParticleContainer& xTruthParticleContainer,
                                     const xAOD::TruthParticle* thePart,
                                     bool& isPrompt,
                                     MCTruthPartClassifier::Info& info) const
{
  ATH_MSG_DEBUG("Executing DefOrigOfNeutrino ");

  const int nuFlav = abs(thePart->pdgId());
  // Find the first copy of this particle stored in the xAOD::TruthParticleContainer (i.e. the particle prior to any interactions)
  const xAOD::TruthParticle* thePriPart = MC::findMatching(xTruthParticleContainer, thePart);
  if (!thePriPart) return NonDefined;
  if (abs(thePriPart->pdgId()) != nuFlav) return NonDefined; // FIXME should this be if (!MC::isSMNeutrino(thePriPart) || abs(thePriPart->pdgId()) != nuFlav) return NonDefined; // (Use MC::isNeutrino if 4th generation neutrinos OK)

  const xAOD::TruthVertex* partOrigVtx = thePriPart->hasProdVtx() ? thePriPart->prodVtx() : nullptr;

  //-- to define neutrino outcome status
  info.particleOutCome = NonInteract;

  if (!partOrigVtx) return NonDefined;

  int numOfParents = partOrigVtx->nIncomingParticles();
  if (numOfParents > 1) ATH_MSG_DEBUG("DefOrigOfNeutrino:: neutrino has more than one parent.");

  const xAOD::TruthParticle* ancestor = MC::findMother(thePriPart);
  info.setMotherProperties(ancestor);
  if (!ancestor) { return NonDefined; } // ancestor is not a nullptr beyond this point

  int ancestorPDG = ancestor->pdgId();
  const xAOD::TruthVertex* ancestorProdVtx = ancestor->hasProdVtx() ? ancestor->prodVtx() : nullptr;

  // Start of method 3 of protecting against loops
  // to resolve Sherpa loop
  bool samePart = (ancestorProdVtx && HepMC::is_same_vertex(ancestorProdVtx,partOrigVtx));
  // End of method 3 of protecting against loops

  if ((abs(ancestorPDG) == nuFlav || MC::isTau(ancestorPDG) || MC::isW(ancestorPDG)) && ancestorProdVtx && !samePart) {
    int pPDG(0);
    const xAOD::TruthParticle* AncestorParent{};
    do {
      pPDG = 0;
      AncestorParent = MC::findMother(ancestor);
      // Start of method 2 of protecting against loops
      // to prevent Sherpa loop
      if (TruthLoopDetectionMethod2(ancestor,AncestorParent)) {
        AncestorParent = ancestor;
        break;
      }
      //
      if (AncestorParent) {
        pPDG = AncestorParent->pdgId(); // FIXME difference in behaviour compared to defOrigOfElectron/Muon pPDG set even if we are in a loop
      }
      // to prevent Sherpa loop
      if (ancestor == AncestorParent) {
        break;
      }
      // End of method 2 of protecting against Sherpa loops
      if (abs(pPDG) == nuFlav || MC::isTau(pPDG) || MC::isW(pPDG) ) {
        // There will be another iteration so set ancestor to AncestorParent
        ancestor = AncestorParent; // AncestorParent is not a nullptr
        info.setMotherProperties(ancestor); // FIXME difference in behaviour compared to MCTruthClassifier::defOrigOfElectron/Muon
      }

    } while ((std::abs(pPDG) == nuFlav || MC::isTau(pPDG) || MC::isW(pPDG)));

    if (std::abs(pPDG) == nuFlav || MC::isTau(pPDG) || MC::isW(pPDG) || MC::isZ(pPDG) || MC::isHiggs(pPDG) ||
        MC::isMSSMHiggs(pPDG) || MC::isHeavyBoson(pPDG) || MC::isTop(pPDG) ||  // MSSM Higgs bosons, Heavy bosons( Z', Z'', W'+)
        std::abs(pPDG) == MC::WBOSON_LRSM || MC::isNeutrinoRH(pPDG) || // Left-right symmetric model WBoson || Right-handed neutrino (Pythia-specific)
        MC::isSUSY(pPDG)) {
      ancestor = AncestorParent; // AncestorParent is not nullptr here
      info.setMotherProperties(ancestor);
    }
  }
  //if ancestor is still nullptr, we have a problem
  if (!ancestor) return NonDefined; // FIXME it should not be possible for ancestor to be nullptr at this point???

  info.setMotherProperties(ancestor);
  ancestorPDG = ancestor->pdgId();
  partOrigVtx = ancestor->decayVtx();
  ancestorProdVtx = ancestor->hasProdVtx() ? ancestor->prodVtx() : nullptr;
  numOfParents = partOrigVtx->nIncomingParticles();
  const int numberOfChildren = partOrigVtx->nOutgoingParticles();

  // Determine decay products
  auto DP = DecayProducts(partOrigVtx);
  const int NumOfPhot = DP.pd(MC::PHOTON);
  const int NumOfquark = DP.apd({MC::DQUARK,MC::UQUARK,MC::SQUARK,MC::CQUARK,MC::BQUARK,MC::TQUARK});
  const int NumOfgluon = DP.apd(MC::GLUON);
  const int NumOfLQ = DP.apd(MC::LEPTOQUARK);
  const int NumOfElNeut = DP.apd(MC::NU_E);
  const int NumOfMuNeut = DP.apd(MC::NU_MU);
  const int NumOfTauNeut = DP.apd(MC::NU_TAU);
  const int NumOfEl = DP.apd(MC::ELECTRON);
  const int NumOfMu = DP.apd(MC::MUON);
  const int NumOfTau = DP.apd(MC::TAU);

  samePart = false;
  for (const auto& aChild: partOrigVtx->particles_out()) {
    if (!aChild) continue;
    if (aChild->pdgId() == ancestorPDG && HepMC::is_same_generator_particle(aChild,ancestor)) {
      samePart = true;
      break;
    }
  }
  // End of section determining decay products

  // Quark weak decay
  if (MC::isQuark(ancestorPDG) && numOfParents == 1 && numberOfChildren == 3 && NumOfquark == 1 && (NumOfEl == 1 || NumOfMu == 1 || NumOfTau == 1)) return QuarkWeakDec;
  if (MC::isTop(ancestorPDG)) return top;

  if (MC::isW(ancestorPDG) && ancestorProdVtx && ancestorProdVtx->nIncomingParticles() != 0) {
    const xAOD::TruthVertex* prodVert = ancestorProdVtx;
    const xAOD::TruthParticle* ptrPart;
    do {
      ptrPart = prodVert->incomingParticle(0); // FIXME just taking the first one
      prodVert = ptrPart->hasProdVtx() ? ptrPart->prodVtx() : nullptr;
    } while (MC::isW(ptrPart) && prodVert);

    if (prodVert && prodVert->nIncomingParticles() == 1) {
      if (abs(ptrPart->pdgId()) == MC::RH_NU_E) return NuREle; // Right-handed NU_E (Pythia-specific)
      if (abs(ptrPart->pdgId()) == MC::RH_NU_MU) return NuRMu; // Right-handed NU_MU (Pythia-specific)
      if (abs(ptrPart->pdgId()) == MC::RH_NU_TAU) return NuRTau; // Right-handed NU_TAU (Pythia-specific)
    }
    return WBoson;
  }
  if (MC::isW(ancestorPDG)) return WBoson;
  if (MC::isZ(ancestorPDG)) return ZBoson;

  //-- Exotics

  // MadGraphPythia ZWW*->lllnulnu or ZWW*->nunulnulnu (don't even know if the latter is generated)
  if (numOfParents == 1 && numberOfChildren > 4 && (MC::isSMQuark(ancestorPDG) || MC::isGluon(ancestorPDG))) {

    const xAOD::TruthParticle* thePartToCheck = thePriPart;
    const xAOD::TruthParticle* theParent = thePriPart->hasProdVtx() ? thePriPart->prodVtx()->incomingParticle(0) : nullptr; // FIXME just taking the first one

    if (MC::isElectron(theParent) && MC::isDecayed(theParent)) { thePartToCheck = theParent; }
    bool isZboson = false;
    bool isWboson = false;
    bool skipnext = false;

    for (unsigned int ipOut = 0; ipOut + 1 < partOrigVtx->nOutgoingParticles(); ++ipOut) {
      const xAOD::TruthParticle* aChild = partOrigVtx->outgoingParticle(ipOut);
      if (!aChild) continue;
      const xAOD::TruthParticle* theNextChild{};
      for (unsigned int ipOut1 = ipOut + 1; ipOut1 < partOrigVtx->nOutgoingParticles(); ipOut1++) {
        theNextChild = partOrigVtx->outgoingParticle(ipOut1);
        if (theNextChild) break;
      }
      if (!theNextChild) continue;

      if (skipnext) {
        skipnext = false;
        continue;
      }

      const int apdgID1 = abs(aChild->pdgId());
      const int apdgID2 = abs(theNextChild->pdgId());
      if (apdgID1 == apdgID2 && MC::isSMNeutrino(apdgID1)) {
        // Zboson
        if (thePartToCheck == aChild || thePartToCheck == theNextChild) {
          isZboson = true;
          break;
        }
        skipnext = true;
      } else if ((apdgID1 == MC::ELECTRON && apdgID2 == MC::NU_E) ||
                 (apdgID1 == MC::NU_E && apdgID2 == MC::ELECTRON) ||
                 (apdgID1 == MC::MUON && apdgID2 == MC::NU_MU) ||
                 (apdgID1 == MC::NU_MU && apdgID2 == MC::MUON) ||
                 (apdgID1 == MC::TAU && apdgID2 == MC::NU_TAU) ||
                 (apdgID1 == MC::NU_TAU && apdgID2 == MC::TAU)
                 ) {
        // WBoson
        if (thePartToCheck == aChild || thePartToCheck == theNextChild) {
          isWboson = true;
          break;
        }
        skipnext = true;
      }
    }
    if (isWboson) return WBoson;
    if (isZboson) return ZBoson;
  }

  if (numOfParents == 2) {
    const int pdg1 = partOrigVtx->incomingParticle(0)->pdgId();
    const int pdg2 = partOrigVtx->incomingParticle(1)->pdgId();
    //--Sherpa Z->nunu
    if ( (numberOfChildren - NumOfquark - NumOfgluon) == 2 && (NumOfElNeut == 2 || NumOfMuNeut == 2 || NumOfTauNeut == 2)) return ZBoson;

    //--Sherpa W->enu ??
    if ((numberOfChildren - NumOfquark - NumOfgluon) == 2 && ((NumOfEl == 1 && NumOfElNeut == 1) || (NumOfMu == 1 && NumOfMuNeut == 1) || (NumOfTau == 1 && NumOfTauNeut == 1))) return WBoson;

    //--Sherpa ZZ,ZW
    if ( (numberOfChildren - NumOfquark - NumOfgluon) == 4 && (NumOfEl + NumOfMu + NumOfTau + NumOfElNeut + NumOfMuNeut + NumOfTauNeut == 4) &&
         (MC::isQuark(pdg1)||MC::isGluon(pdg1)) && (MC::isQuark(pdg2)||MC::isGluon(pdg2))) return DiBoson;

    //--Sherpa VVV -- Note, have to allow for prompt photon radiation or these get lost
    if ((numberOfChildren - NumOfquark - NumOfgluon - NumOfPhot) == 6 && (NumOfEl + NumOfMu + NumOfTau + NumOfElNeut + NumOfMuNeut + NumOfTauNeut == 6) &&
        (MC::isQuark(pdg1)||MC::isGluon(pdg1)) && (MC::isQuark(pdg2)||MC::isGluon(pdg2))) return MultiBoson;
  }

  // New Sherpa Z->nunu
  if (partOrigVtx == ancestorProdVtx && partOrigVtx) {
    int NumOfLepLoop = 0;
    int NumOfNeuLoop = 0;
    for (const auto *const pout: partOrigVtx->particles_out()) {
      if (!pout) continue;
      for (const auto *const pin: partOrigVtx->particles_in()) {
        if (!pin) continue;
        if (HepMC::is_same_particle(pin,pout)) continue;
        const int apdgid = abs(pout->pdgId());
        if (MC::isSMLepton(apdgid)) {
          if (MC::isSMNeutrino(apdgid)) { NumOfNeuLoop++; }
          else { NumOfLepLoop++; }
        }
      }
    }
    if (NumOfNeuLoop == 2 && NumOfLepLoop == 0) return ZBoson;
    if (NumOfNeuLoop == 1 && NumOfLepLoop == 1) return WBoson;
    if (NumOfNeuLoop + NumOfLepLoop == 4) return DiBoson;
  }

  //-- McAtNLo

  if (MC::isHiggs(ancestorPDG)) return Higgs;
  if (MC::isMSSMHiggs(ancestorPDG)) return HiggsMSSM; // MSSM Higgs bosons
  if (MC::isHeavyBoson(ancestorPDG)) return HeavyBoson; // Heavy bosons( Z', Z'', W'+)

  if (MC::isTau(ancestorPDG)) {
    const ParticleOrigin tauOrig = defOrigOfTau(xTruthParticleContainer, ancestor, ancestorPDG, info);
    const ParticleType tautype = defTypeOfTau(tauOrig);
    return (tautype == IsoTau)?tauOrig:TauLep;
  }

  if (abs(ancestorPDG) == MC::WBOSON_LRSM) return WBosonLRSM; // Left-right symmetric model WBoson (Pythia-specific)
  if (abs(ancestorPDG) == MC::RH_NU_E) return NuREle; // Right-handed NU_E (Pythia-specific)
  if (abs(ancestorPDG) == MC::RH_NU_MU) return NuRMu; // Right-handed NU_MU (Pythia-specific)
  if (abs(ancestorPDG) == MC::RH_NU_TAU) return NuRTau; // Right-handed NU_TAU (Pythia-specific)
  if (MC::isLeptoQuark(ancestorPDG) || NumOfLQ != 0) return LQ;
  if (MC::isSUSY(ancestorPDG)) return SUSY;
  if (MC::isBSM(ancestorPDG)) return OtherBSM;

  const ParticleType pType = defTypeOfHadron(ancestorPDG);
  if ((pType == BBbarMesonPart || pType == CCbarMesonPart) && ancestorProdVtx && MC::isHardScatteringVertex(ancestorProdVtx)) isPrompt = true;

  return convHadronTypeToOrig(pType, ancestorPDG);
}
