/*
  Copyright (C) 2002-2025 CERN for the benefit of the ATLAS collaboration
*/

/// @author Nils Krumnack



//
// includes
//

#include "CxxUtils/checker_macros.h"
ATLAS_NO_CHECK_FILE_THREAD_SAFETY;

#include <AsgTesting/UnitTest.h>
#include <AsgTools/AsgToolConfig.h>
#include <ColumnarTestFixtures/ColumnarMemoryTest.h>
#include <ColumnarTestFixtures/ColumnarPhysliteTest.h>
#include <xAODCore/ShallowCopy.h>

#include <MuonEfficiencyCorrections/MuonEfficiencyScaleFactors.h>
#include <xAODEventInfo/EventInfo.h>
#include <xAODMuon/Muon.h>

#pragma GCC diagnostic ignored "-Wdeprecated-declarations"

//
// method implementations
//

using columnar::ColumnarMemoryTest;
using columnar::ColumnarPhysLiteTest;

TEST_F (ColumnarMemoryTest, MuonEfficiencyScaleFactors)
{
  if (!checkMode())
    return;

  asg::AsgToolConfig toolConfig;
  toolConfig.setTypeAndName ("CP::MuonEfficiencyScaleFactors/" + makeUniqueName());
  ToolHandle<asg::AsgTool> myToolHandle;
  std::shared_ptr<void> cleanup;
  ASSERT_SUCCESS (toolConfig.makeTool (myToolHandle, cleanup));
  ColumnarTestToolHandle toolHandle (*myToolHandle);
  toolHandle.initialize ();

  for (auto& name : toolHandle.getColumnNames())
    std::cout << "requested column: " << name << std::endl;
  std::cout << "recommended systematics size: " << toolHandle.getRecommendedSystematics().size() << std::endl;

  ColumnMapType columnMap {toolHandle};

  columnMap.addColumn ("EventInfo", {0, 1});
  columnMap.addColumn ("EventInfo.runNumber", {284500}); // no specific reason, from ASG test file
  columnMap.addColumn ("EventInfo.RandomRunNumber", {284500}); // copy from above
  columnMap.addColumn ("EventInfo.eventTypeBitmask", {unsigned(xAOD::EventInfo::IS_SIMULATION)});

  columnMap.addColumn ("Muons", {0, 1});
  columnMap.addColumn ("Muons.pt", {10e5});
  columnMap.addColumn ("Muons.eta", {1});
  columnMap.addColumn ("Muons.phi", {1});
  columnMap.addColumn ("Muons.muonType", {unsigned(xAOD::Muon::MuonType::Combined)});

  columnMap.addColumn ("Muons.sfOut", {0});
  columnMap.addColumn ("Muons.validOut", {0});

  // there is no special reason for this value, it is just what came
  // out of my first test run.  if the tool changes, feel free to
  // update this value.
  columnMap.setExpectation ("Muons.sfOut", {0.99509060382843018});
  columnMap.setExpectation ("Muons.validOut", {1});

  columnMap.connectColumnsToTool ();

  columnMap.call ();

  columnMap.checkExpectations ();
}


TEST_F (ColumnarMemoryTest, MuonEfficiencyScaleFactors_systematic)
{
  if (!checkMode())
    return;

  asg::AsgToolConfig toolConfig;
  toolConfig.setTypeAndName ("CP::MuonEfficiencyScaleFactors/" + makeUniqueName());
  ToolHandle<asg::AsgTool> myToolHandle;
  std::shared_ptr<void> cleanup;
  ASSERT_SUCCESS (toolConfig.makeTool (myToolHandle, cleanup));
  ColumnarTestToolHandle toolHandle (*myToolHandle);
  for (auto& systematic : toolHandle.getRecommendedSystematics())
    toolHandle.applySystematicVariation (systematic);
  toolHandle.initialize ();

  for (auto& name : toolHandle.getColumnNames())
    std::cout << "requested column: " << name << std::endl;
  std::cout << "recommended systematics size: " << toolHandle.getRecommendedSystematics().size() << std::endl;

  for (auto& systematic : toolHandle.getRecommendedSystematics())
  {
    std::cout << "systematic: " << systematic << std::endl;

    toolHandle.applySystematicVariation (systematic);

    ColumnMapType columnMap {toolHandle};

    columnMap.addColumn ("EventInfo", {0, 1});
    columnMap.addColumn ("EventInfo.runNumber", {284500}); // no specific reason, from ASG test file
    columnMap.addColumn ("EventInfo.RandomRunNumber", {284500}); // copy from above
    columnMap.addColumn ("EventInfo.eventTypeBitmask", {unsigned(xAOD::EventInfo::IS_SIMULATION)});

    columnMap.addColumn ("Muons", {0, 1});
    columnMap.addColumn ("Muons.pt", {10e5});
    columnMap.addColumn ("Muons.eta", {1});
    columnMap.addColumn ("Muons.phi", {1});
    columnMap.addColumn ("Muons.muonType", {unsigned(xAOD::Muon::MuonType::Combined)});

    columnMap.addColumn ("Muons.sfOut", {0});
    columnMap.addColumn ("Muons.validOut", {0});

    columnMap.connectColumnsToTool ();

    columnMap.call ();
  }
}



TEST_F (ColumnarMemoryTest, MuonEfficiencyScaleFactors_multiEvent)
{
  if (!checkMode())
    return;

  asg::AsgToolConfig toolConfig;
  toolConfig.setTypeAndName ("CP::MuonEfficiencyScaleFactors/" + makeUniqueName());
  ToolHandle<asg::AsgTool> myToolHandle;
  std::shared_ptr<void> cleanup;
  ASSERT_SUCCESS (toolConfig.makeTool (myToolHandle, cleanup));
  ColumnarTestToolHandle toolHandle (*myToolHandle);
  toolHandle.initialize ();

  for (auto& name : toolHandle.getColumnNames())
    std::cout << "requested column: " << name << std::endl;
  std::cout << "recommended systematics size: " << toolHandle.getRecommendedSystematics().size() << std::endl;

  ColumnMapType columnMap {toolHandle};

  columnMap.addColumn ("EventInfo", {0, 200});
  columnMap.addColumn ("EventInfo.runNumber", {310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000, 310000}); // no specific reason, from ASG test file
  columnMap.addColumn ("EventInfo.RandomRunNumber", {363262, 357539, 364076, 351894, 357451, 357821, 360309, 350479, 350160, 352340, 357962, 350923, 349637, 351698, 362445, 358395, 360129, 360402, 349646, 358656, 363198, 352436, 358096, 362552, 352448, 358333, 358115, 349842, 350144, 358395, 357409, 363262, 351894, 359918, 349842, 355563, 351223, 356250, 349114, 364076, 359124, 352448, 358395, 363033, 350184, 359872, 359355, 363129, 360293, 349335, 356259, 350160, 358233, 360293, 352340, 361862, 359872, 356124, 350880, 349481, 360209, 352448, 349114, 364292, 358031, 350531, 359441, 360402, 363033, 355861, 359872, 351969, 364098, 363198, 350184, 358577, 363710, 350682, 357772, 350682, 362345, 352448, 358233, 363198, 359541, 358395, 349693, 360373, 356124, 356259, 361862, 350440, 350923, 360209, 357409, 359191, 356259, 358577, 362661, 363198, 350880, 360293, 349268, 358541, 349977, 362204, 362204, 356250, 352340, 349051, 358395, 356205, 362619, 349111, 358031, 350479, 362204, 359191, 350848, 351364, 351364, 358031, 350013, 356177, 360161, 355563, 361738, 352448, 350361, 350848, 352394, 357962, 358333, 349268, 359171, 355754, 357962, 363979, 362388, 355261, 351364, 349169, 357620, 358215, 362552, 364098, 364292, 363710, 361795, 350144, 359823, 359872, 350160, 350479, 357451, 351296, 349335, 358541, 355599, 349944, 364030, 350184, 363129, 359191, 349533, 364214, 351455, 351628, 351325, 352056, 351223, 356250, 363262, 351062, 350144, 363400, 349944, 357713, 355995, 349498, 358300, 352340, 358395, 358115, 355651, 351628, 360293, 357821, 349944, 358516, 359823, 352056, 357679, 358541, 350531, 361795, 364214, 350013, 350160, 359124}); // copy from above
  columnMap.addColumn ("EventInfo.eventTypeBitmask", {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1});

  columnMap.addColumn ("Muons", {0, 0, 1, 2, 2, 2, 5, 5, 7, 7, 8, 10, 10, 10, 10, 11, 12, 12, 13, 14, 16, 17, 17, 18, 19, 22, 22, 22, 24, 24, 24, 24, 25, 25, 25, 25, 26, 26, 27, 28, 28, 28, 29, 29, 30, 31, 31, 32, 33, 33, 34, 34, 35, 35, 36, 36, 36, 37, 37, 39, 40, 43, 44, 45, 45, 45, 45, 47, 48, 49, 53, 54, 55, 57, 58, 58, 58, 60, 60, 60, 61, 63, 64, 64, 64, 64, 64, 65, 65, 67, 68, 71, 72, 72, 72, 75, 76, 76, 76, 78, 79, 79, 79, 79, 80, 80, 81, 81, 81, 81, 83, 83, 85, 86, 87, 87, 88, 88, 89, 89, 91, 91, 93, 95, 96, 96, 97, 99, 101, 102, 102, 103, 103, 103, 105, 105, 106, 108, 108, 108, 108, 109, 111, 112, 113, 113, 115, 117, 117, 118, 119, 122, 125, 126, 127, 127, 128, 129, 129, 131, 131, 132, 133, 134, 136, 137, 140, 141, 143, 143, 144, 144, 145, 146, 148, 148, 148, 148, 148, 150, 150, 151, 152, 154, 155, 156, 157, 160, 160, 160, 161, 161, 161, 161, 163, 166, 167, 168, 169, 170, 170});
  columnMap.addColumn ("Muons.pt", {44734.3515625, 5436.24609375, 22758.580078125, 9240.9677734375, 6871.32275390625, 71743.1015625, 4113.740234375, 61487.10546875, 27169.630859375, 8629.369140625, 8690.1328125, 25952.08203125, 43775.51953125, 13358.6796875, 8930.3427734375, 3746.462158203125, 61488.2578125, 11634.1220703125, 3627.313720703125, 74599.9453125, 6948.33544921875, 5926.94677734375, 43564.97265625, 27756.92578125, 79462.65625, 13855.875, 7055.65869140625, 47684.0390625, 4838.04052734375, 3807.7626953125, 141233.421875, 9890.85546875, 55530.2421875, 16722.197265625, 16626.88671875, 26141.869140625, 67852.6484375, 28505.998046875, 7928.3994140625, 44053.38671875, 12651.310546875, 12461.2158203125, 4447.88037109375, 5625.02099609375, 20851.826171875, 45458.58984375, 21059.3046875, 19216.41796875, 11349.794921875, 50556.31640625, 38504.18359375, 16806.529296875, 5488.41650390625, 47570.89453125, 29549.86328125, 75017.53125, 4792.79443359375, 17203.69140625, 47059.671875, 19696.392578125, 4370.5263671875, 40759.546875, 26184.390625, 90218.5, 14980.9892578125, 44169.3046875, 3262.626708984375, 19839.755859375, 42028.33203125, 21627.29296875, 6204.08056640625, 4545.80126953125, 4906.5498046875, 4468.07763671875, 3569.943115234375, 52666.66015625, 75249.203125, 5237.369140625, 31833.16015625, 38780.87890625, 21972.91015625, 22945.4140625, 9364.70703125, 11701.3193359375, 8788.5888671875, 36886.6015625, 4472.42041015625, 10758.607421875, 4834.30126953125, 43464.65234375, 4402.01220703125, 6208.49609375, 5702.42431640625, 33272.38671875, 8839.2939453125, 84198.953125, 18292.138671875, 133491.546875, 12369.7705078125, 39570.0, 32628.080078125, 117191.8359375, 4092.18798828125, 62641.734375, 24193.40234375, 4928.0322265625, 59307.53515625, 40508.3359375, 68577.1796875, 23650.880859375, 11044.3291015625, 36658.5859375, 62960.79296875, 57539.40625, 3799.839111328125, 54333.26953125, 40482.25, 10121.1728515625, 10163.568359375, 135374.484375, 7481.8232421875, 4650.67529296875, 35063.76171875, 29179.30078125, 3381.900390625, 27647.35546875, 87986.4375, 23238.943359375, 47818.12890625, 21002.2890625, 6677.42236328125, 10359.171875, 4427.748046875, 72032.8828125, 68145.46875, 6380.78662109375, 18389.65234375, 90672.71875, 54056.30859375, 40142.66015625, 7114.48486328125, 52966.7109375, 6781.369140625, 66765.8671875, 11932.0478515625, 9232.525390625, 27730.8359375, 20614.84765625, 52352.1953125, 36006.8359375, 20066.5, 36421.1015625, 28618.9296875, 27087.91015625, 5306.1572265625, 137598.71875, 7079.8701171875, 38577.22265625, 20152.091796875, 9841.009765625, 41422.5703125, 19459.26171875, 4849.69873046875, 29555.20703125, 20169.162109375, 16237.001953125, 11802.7265625, 51601.64453125, 4993.3759765625, 18289.650390625});
  columnMap.addColumn ("Muons.eta", {-1.3528014421463013, -2.135263204574585, -0.6826117634773254, -1.3291620016098022, -1.8656655550003052, 1.8440676927566528, 1.9415327310562134, 1.2437396049499512, 0.8164210915565491, -1.211382269859314, -1.7200701236724854, -0.1586022824048996, -0.8991519808769226, -1.3649072647094727, -0.7469519972801208, 0.047038085758686066, 0.9873219132423401, 1.361953616142273, -0.21793636679649353, -0.5423484444618225, 1.1687898635864258, -1.37875497341156, 1.3138712644577026, 0.6487455368041992, -0.09575893729925156, 0.7963840961456299, -0.5974398255348206, 1.0957417488098145, -1.2813899517059326, -1.5148266553878784, -2.291321039199829, 2.1897995471954346, -2.0790791511535645, -0.2230755239725113, 0.32960304617881775, -2.0772225856781006, 1.48035728931427, -1.326144814491272, 0.21418103575706482, -0.7924028635025024, -1.6496127843856812, 0.06894388794898987, -0.1325385868549347, -0.8035625219345093, -1.511734127998352, 0.49724575877189636, -2.3305718898773193, 0.7735245227813721, -2.3323519229888916, -0.1627490073442459, 0.21484065055847168, -0.21887771785259247, 1.6168524026870728, -2.5414247512817383, 2.165271043777466, -0.27472424507141113, 0.9847943782806396, 0.7965577244758606, 0.9277859330177307, 1.5038728713989258, 0.39742547273635864, -1.3974103927612305, -0.03107849881052971, 1.1376224756240845, 0.34766560792922974, 0.06560862809419632, -0.9155166149139404, 0.5450497269630432, -1.1272151470184326, -0.6064988374710083, 0.30741316080093384, -1.7032257318496704, -0.33583998680114746, -0.3121020495891571, -0.3685411810874939, 2.042809009552002, 0.332103967666626, 0.038575515151023865, -0.9359152913093567, 0.5225610136985779, -1.2370107173919678, -0.0723504051566124, 0.7575925588607788, -0.28907933831214905, 0.13540683686733246, -1.8538968563079834, -1.5179120302200317, 1.6840845346450806, 1.9197245836257935, -0.9007341861724854, -1.3331900835037231, 0.9711335897445679, 1.0146195888519287, -1.5946853160858154, -2.45310640335083, 2.4087820053100586, 0.7294056415557861, 1.739258050918579, 1.3442418575286865, 2.1322813034057617, 1.335025668144226, -0.08004572242498398, -1.7912588119506836, 0.5233050584793091, -1.6048105955123901, 0.2171621471643448, -0.5036993622779846, 0.6023936867713928, -0.5958006978034973, -1.9275152683258057, 1.046723484992981, 0.721760630607605, -0.39625900983810425, -1.883686900138855, 1.358102798461914, 0.5024107694625854, 0.5202184319496155, -0.564322829246521, -2.0200278759002686, 2.183765172958374, 0.19610309600830078, 0.03482528403401375, 0.2452843189239502, 0.23854801058769226, 0.2941405773162842, -1.4226300716400146, -1.7840321063995361, 1.4091660976409912, -0.8756387829780579, 0.6649539470672607, 0.8186972737312317, 0.32528170943260193, -0.7350008487701416, 1.503635287284851, 0.5185919404029846, 1.0362918376922607, 0.19513817131519318, 1.013355016708374, 0.46696802973747253, 1.0780260562896729, -0.2102329134941101, -0.7191573977470398, -0.5705442428588867, 2.6653006076812744, 0.7459214925765991, -0.7356237769126892, 0.20598091185092926, 0.5172070860862732, 0.09819929301738739, -0.3523336946964264, -1.4815741777420044, 0.3791424334049225, 0.7322942018508911, -1.8588688373565674, -1.1228258609771729, 0.6478890180587769, -1.7408149242401123, -2.0416510105133057, -2.04489803314209, -2.1347498893737793, 0.38880491256713867, -1.324798345565796, -0.23969298601150513, -1.9375287294387817, -1.154435396194458, -0.6551217436790466, 0.19739054143428802, -0.7563914656639099, -0.026105551049113274, -0.43444475531578064});
  columnMap.addColumn ("Muons.phi", {-0.42633694410324097, -1.118241548538208, -0.2937942445278168, -1.8621268272399902, 1.3888180255889893, 1.547302484512329, -1.3970820903778076, -1.4286469221115112, 2.113107442855835, -2.688753843307495, -0.528590738773346, 1.305529236793518, 0.40013378858566284, 2.8415353298187256, 0.6112218499183655, 0.04430275782942772, -2.376418113708496, 0.009483779780566692, 0.4926173686981201, -2.9813876152038574, 0.19588780403137207, 1.496896505355835, 2.474069356918335, 1.0449435710906982, 2.4909942150115967, 2.7024588584899902, 2.3073348999023438, 2.705040693283081, -0.9677503108978271, -2.9008829593658447, -2.2080116271972656, 0.23455286026000977, -1.0964857339859009, -1.4798513650894165, 0.8190762400627136, 0.5219516754150391, 0.7315099835395813, 2.731565475463867, -2.9373867511749268, 1.232192039489746, -1.213453769683838, 1.0778768062591553, -0.014137056656181812, -2.4795188903808594, 1.0242024660110474, -3.0889666080474854, -0.6184263229370117, 2.9734604358673096, 0.3219912052154541, -1.0037897825241089, -0.09857949614524841, 2.744572639465332, -1.9659918546676636, -1.1343016624450684, -1.5744571685791016, 2.5157010555267334, -2.5180914402008057, -2.7506463527679443, -1.7956726551055908, 0.9136759042739868, -2.8834128379821777, 1.257405400276184, -0.6138610243797302, 1.9350438117980957, -0.4064626097679138, 0.7196569442749023, -1.7367417812347412, -2.9993507862091064, 1.7891151905059814, 1.4899842739105225, 0.10527615249156952, -0.8888806700706482, 1.4110093116760254, 1.2597523927688599, 1.1587697267532349, -2.7394354343414307, -2.7208359241485596, 1.0396074056625366, -0.8221499919891357, -0.3988962173461914, 0.7699745893478394, 0.5492132902145386, -0.6982453465461731, -2.887334108352661, -1.8004684448242188, -1.2792845964431763, 0.790737509727478, -1.2367711067199707, 0.6265870928764343, 2.1759822368621826, 2.228562116622925, -1.3034783601760864, 2.038651943206787, 2.9271130561828613, 2.4882583618164062, 2.682018280029297, -2.0639331340789795, 1.8784698247909546, 2.230875253677368, 2.19922137260437, 0.1794716715812683, -2.0462687015533447, -2.7756879329681396, -1.7162082195281982, 1.0460305213928223, 2.1269826889038086, -1.2233872413635254, -1.0923856496810913, -0.4357268810272217, -2.31498384475708, -1.936191201210022, 0.809776246547699, -1.0558210611343384, -2.940261125564575, -0.0930144339799881, -0.8736723065376282, -2.895345449447632, -3.1259186267852783, 1.6756242513656616, -2.795151948928833, 0.3850914239883423, 0.35255104303359985, 1.3704909086227417, 2.0322999954223633, 1.8855817317962646, -1.3853029012680054, -1.837311863899231, -0.04169650375843048, -0.7792194485664368, 2.7664124965667725, 1.1724512577056885, -0.30960679054260254, -2.0559194087982178, -3.0110130310058594, 1.9376403093338013, -2.5238091945648193, 0.02127033844590187, -1.0474599599838257, -0.45513880252838135, 3.000321626663208, 2.676793098449707, 2.620002508163452, 0.016491053625941277, -1.0034220218658447, -0.5219187140464783, 2.41656494140625, 2.2668333053588867, 2.683040142059326, 0.8519139885902405, 0.8576232194900513, 2.5523300170898438, -2.355067014694214, -2.2060999870300293, 0.8053587079048157, 1.0153076648712158, -2.525090217590332, -0.9989657402038574, -1.7044318914413452, -1.6776373386383057, 2.1262362003326416, -1.7768948078155518, 3.094853401184082, 0.04889829829335213, 1.0182515382766724, 1.183576226234436, -1.0860525369644165, 1.6805686950683594, -0.2419508546590805, -1.8496487140655518, -1.6465376615524292});
  columnMap.addColumn ("Muons.muonType", {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0});

  columnMap.addColumn ("Muons.sfOut", {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0});
  columnMap.addColumn ("Muons.validOut", {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0});

  // there is no special reason for this value, it is just what came
  // out of my first test run.  if the tool changes, feel free to
  // update this value.
  columnMap.setExpectation ("Muons.sfOut", {0.99831742, 1.0046741, 0.99739212, 1.0004433, 1.0046754, 0.99716991, 0.99509174, 0.9989965, 0.99633616, 1.0025316, 1.0004433, 0.99852306, 0.99627501, 1.0015301, 1.0032424, 1.0642542, 0.99910098, 0.99947166, 1.0536422, 0.99842858, 1.0070994, 1.0027269, 0.99807477, 0.99829549, 0.89542562, 1.0025811, 1.0062463, 0.99686581, 1.0861259, 1.0220662, 0.99617219, 0.99431932, 0.99658382, 0.99765778, 0.99841005, 0.99538261, 0.99796283, 0.99855262, 1.0062758, 0.99628937, 1.0015301, 1.3279924, 1.0570323, 1.02034, 0.99891704, 0.99912053, 0.99683291, 0.99848717, 1.0006404, 0.99500179, 0.9979102, 0.99850231, 1.0015724, 0.98250622, 0.97950631, 1.0016543, 1.0735306, 0.99673402, 0.99269658, 0.99796283, 1.0735306, 0.99891704, 1.0017905, 0.99393636, 1.0025811, 0.99667937, 0.81128389, 0.99912053, 0.99482864, 0.99825573, 1.0105151, 1.0106881, 1.0570323, 1.0570323, 1.0536422, 0.99617648, 0.99848658, 1.1922423, 0.99905288, 0.9971689, 0.99907684, 0.99715298, 1.0041881, 1.0031132, 1.0041881, 0.99747038, 1.0106881, 0.99947166, 0.99509174, 0.99944186, 1.0106881, 1.0105151, 1.0180148, 0.99855262, 0.9988454, 0.99692416, 0.99269658, 0.99393129, 0.99931568, 0.99220508, 0.9983753, 0.98754454, 1.0106881, 0.99809504, 0.99776292, 1.0735306, 0.99442071, 0.99376303, 0.99739212, 0.99704981, 1.0014209, 0.99900186, 0.99500179, 0.99671555, 1.021626, 0.99838251, 0.99729806, 1.0031132, 1.0006404, 0.99617648, 1.0062758, 1.1117166, 0.99902046, 0.99808437, 0.80155939, 0.99810392, 0.99773902, 0.9983753, 0.99905288, 0.99734235, 1.0105151, 1.0014209, 1.0570323, 0.99731308, 0.99794006, 1.0105151, 0.9979102, 0.99513191, 0.9971689, 0.99807119, 1.0062463, 0.99612647, 1.01078, 0.84819293, 1.0014209, 1.0032424, 0.99566734, 0.99734235, 0.99667937, 0.99809688, 0.99745262, 0.99802738, 0.99910098, 0.99753612, 1.0183904, 0.99883401, 1.0008529, 0.99639571, 0.99639571, 0.9988454, 0.99847144, 0.99804926, 1.0570323, 0.99776292, 0.99648851, 0.99442071, 1.0014209, 0.99577194, 1.1117166, 0.99212652});
  columnMap.setExpectation ("Muons.validOut", {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1});
  columnMap.connectColumnsToTool ();

  columnMap.call ();

  columnMap.checkExpectations ();
}

void callXAOD (const CP::MuonEfficiencyScaleFactors& tool, bool isPrepCall, const std::string& name) {
  using namespace asg::msgUserCode;
  if (isPrepCall)
  {
      const xAOD::MuonContainer *muons = nullptr;
      ANA_CHECK_THROW (tool.evtStore()->retrieve (muons, name));
      auto [muonsCopy, auxCopy] = xAOD::shallowCopyContainer (*muons);
      const xAOD::EventInfo *eventInfo = nullptr;
      ANA_CHECK_THROW (tool.evtStore()->retrieve (eventInfo, "EventInfo"));
      tool.callSingleEvent (*muonsCopy, *eventInfo);
      delete muonsCopy;
      delete auxCopy;
  } else
  {
      const xAOD::MuonContainer *muons = nullptr;
      ANA_CHECK_THROW (tool.evtStore()->retrieve (muons, name));
      const xAOD::EventInfo *eventInfo = nullptr;
      ANA_CHECK_THROW (tool.evtStore()->retrieve (eventInfo, "EventInfo"));
      tool.callSingleEvent (*muons, *eventInfo);
  }
}


TEST_F (ColumnarPhysLiteTest, MuonEfficiencyScaleFactors)
{
  asg::AsgToolConfig toolConfig;
  toolConfig.setTypeAndName ("CP::MuonEfficiencyScaleFactors/" + makeUniqueName());
  ToolHandle<CP::MuonEfficiencyScaleFactors> myToolHandle;
  std::shared_ptr<void> cleanup;
  ASSERT_SUCCESS (toolConfig.makeTool (myToolHandle, cleanup));

  doCall (*myToolHandle, "MuonEfficiencyScaleFactors", "AnalysisMuons", [&] (auto& args) {callXAOD (*myToolHandle, args.isPrepCall, args.inputContainer);}, {{"Muons", "AnalysisMuons"}});
}


TEST_F (ColumnarPhysLiteTest, MuonEfficiencyScaleFactors_systematics)
{
  asg::AsgToolConfig toolConfig;
  toolConfig.setTypeAndName ("CP::MuonEfficiencyScaleFactors/" + makeUniqueName());
  ToolHandle<CP::MuonEfficiencyScaleFactors> myToolHandle;
  std::shared_ptr<void> cleanup;
  ASSERT_SUCCESS (toolConfig.makeTool (myToolHandle, cleanup));

  doCall (*myToolHandle, "MuonEfficiencyScaleFactors", "AnalysisMuons", [&] (auto& args) {callXAOD (*myToolHandle, args.isPrepCall, args.inputContainer);}, {{"Muons", "AnalysisMuons"}}, "MUON_EFF_RECO_STAT__1down");
}

ATLAS_GOOGLE_TEST_MAIN
