# Copyright (C) 2002-2025 CERN for the benefit of the ATLAS collaboration

import string

class FeatureHandler:
    
    def getFeatureNames(self):
        listFeatNames = []
        for iFeature in self.m_Feature_Names:
            listFeatNames += [self.m_Feature_Names[iFeature]]
        return listFeatNames
    #end getFeatureNames
    
    def getFeatureTypes(self):
        listFeatTypes = []
        for iFeature in self.m_Feature_Types:
            listFeatTypes += [self.m_Feature_Types[iFeature]]
        return listFeatTypes
    #end getFeatureTypes
    
    def getFeatureDefaultValues(self):
        listFeatDefaults = []
        for iFeature in self.m_Feature_Defaults:
            listFeatDefaults += [self.m_Feature_Defaults[iFeature]]
        return listFeatDefaults
    #end getFeatureDefaultValues
    
    def addToFeatures(self, theBaseName, theVarTypeName, theFeatureType, defaultValueOverride=0):
        """ Add a single feature with given name, type and default value """
        
        #set feature name and default value
        featureName     = theVarTypeName + "_" + theBaseName
        defaultValue    = self.m_DefaultValues[theVarTypeName]
        if defaultValueOverride != 0:
            defaultValue    = defaultValueOverride
        
        #add to features
        self.m_Feature_Names[featureName]    = featureName
        self.m_Feature_Defaults[featureName] = defaultValue
        self.m_Feature_Types[featureName]    = theFeatureType
    #end addToFeatures
    
    def addToFeatures_FullName(self, completeFeatureName, theVarTypeName, theFeatureType, defaultValueOverride=0):
        #set feature name and default value
        featureName     = completeFeatureName
        defaultValue    = self.m_DefaultValues[theVarTypeName]
        if defaultValueOverride != 0:
            defaultValue    = defaultValueOverride
        
        #add to features
        self.m_Feature_Names[featureName]    = featureName
        self.m_Feature_Defaults[featureName] = defaultValue
        self.m_Feature_Types[featureName]    = theFeatureType
    #end addToFeatures_FullName
    
    
    def addToFeatures_AllTypes(self, theBaseName, theVarTypeName, theFeatureType, defaultValueOverride=0):
        """ Add the feature described by theName, theType and theDefault for all known constituent types"""
        
        #set default value
        defaultValue    = self.m_DefaultValues[theVarTypeName]
        if defaultValueOverride != 0:
            defaultValue = defaultValueOverride
        
        #fill feature for all types
        for iConstType in self.m_ConstituentTypes:
            featureName     = iConstType + "_" + theVarTypeName + "_" + theBaseName
            
            self.m_Feature_Names[featureName]    = featureName
            self.m_Feature_Defaults[featureName] = defaultValue
            self.m_Feature_Types[featureName]    = theFeatureType
        #end loop over types
    #end addToFeatures_AllTypes
    
    
    def addToFeatures_AllTypes_AllEnergyVariants(self, theBaseName, theVarTypeName, theFeatureType, defaultValueOverride=0):
        """ Add the feature described by theName, theType and theDefault for all known constituent types"""
        for iEnergyType in self.m_EnergyVariantsList:
            featureName     = theBaseName + iEnergyType
            self.addToFeatures_AllTypes(featureName, theVarTypeName, theFeatureType, defaultValueOverride)
        #end loop over energy types
    #end addToFeatures_AllTypes
    
    def __init__(self):
        
        #setup the varType prefixes
        # they are passed to the feature extractor later on - avoids errors due to typos
        self.m_VarTypeName_Sum          = "Sum"
        self.m_VarTypeName_Ratio        = "Ratio"
        self.m_VarTypeName_EtInRing     = "EtInRing"
        self.m_VarTypeName_Isolation    = "Isolation"
        self.m_VarTypeName_Num          = "Num"
        self.m_VarTypeName_Mean         = "Mean"
        self.m_VarTypeName_StdDev       = "StdDev"
        self.m_VarTypeName_HLV          = "HLV"
        self.m_VarTypeName_Angle        = "Angle"
        self.m_VarTypeName_DeltaR       = "DeltaR"
        self.m_VarTypeName_JetMoment    = "JetMoment"
        self.m_VarTypeName_Combined     = "Combined"
        self.m_VarTypeName_JetShape     = "JetShape"
        self.m_VarTypeName_ImpactParams = "ImpactParams"
        self.m_VarTypeName_Basic        = "Basic"
        self.m_VarTypeName_PID          = "PID"
        self.m_VarTypeName_Shots        = "Shots"
        
        #setup the default values for the different variable types
        self.m_DefaultValues = {}
        self.m_DefaultValues[self.m_VarTypeName_Sum]          = -4000.0
        self.m_DefaultValues[self.m_VarTypeName_Ratio]        = -0.2
        self.m_DefaultValues[self.m_VarTypeName_EtInRing]     = -1000.0
        self.m_DefaultValues[self.m_VarTypeName_Isolation]    = -0.2
        self.m_DefaultValues[self.m_VarTypeName_Num]          = -5.0
        self.m_DefaultValues[self.m_VarTypeName_Mean]         = -0.2
        self.m_DefaultValues[self.m_VarTypeName_StdDev]       = -0.2
        self.m_DefaultValues[self.m_VarTypeName_HLV]          = -4000.0
        self.m_DefaultValues[self.m_VarTypeName_Angle]        = -0.2
        self.m_DefaultValues[self.m_VarTypeName_DeltaR]       = -0.2
        self.m_DefaultValues[self.m_VarTypeName_JetMoment]    = -0.2
        self.m_DefaultValues[self.m_VarTypeName_Combined]     = -10.0
        self.m_DefaultValues[self.m_VarTypeName_JetShape]     = -2.0
        self.m_DefaultValues[self.m_VarTypeName_ImpactParams] = -100.
        self.m_DefaultValues[self.m_VarTypeName_Basic]        = -5.0
        self.m_DefaultValues[self.m_VarTypeName_PID]          = -9.0
        self.m_DefaultValues[self.m_VarTypeName_Shots]        = -2.0
        
        #setup the constituent types
        self.m_ConstituentTypeName_All      = "All"
        self.m_ConstituentTypeName_Charged  = "Charged"
        self.m_ConstituentTypeName_Neutral  = "Neutral"
        self.m_ConstituentTypeName_Pi0Neut  = "Pi0Neut"
        self.m_ConstituentTypeName_OutNeut  = "OuterNeut"
        self.m_ConstituentTypeName_OutChrg  = "OuterChrg"
        
        self.m_ConstituentTypes = []
        #baseline
        self.m_ConstituentTypes += [self.m_ConstituentTypeName_Charged]
        self.m_ConstituentTypes += [self.m_ConstituentTypeName_Neutral]
        self.m_ConstituentTypes += [self.m_ConstituentTypeName_Pi0Neut]
        
        self.m_EnergyVariantsList = []
        self.m_EnergyVariantsList += ["EtAllConsts"]
        # ==============================================
        
        
        #setup member variables to hold feature info
        self.m_Feature_Names      = {}
        self.m_Feature_Defaults   = {}
        self.m_Feature_Types      = {}
        
        #fill member variables
        self.addSingleVariables()
        self.addMultiplicities()
        self.addFourMomentum()
        self.addTypeSpecificFeatures_PID()
        self.addTypeSpecificFeatures_Shots()
        self.addTypeSpecificFeatures_Ratios()
        self.addTypeSpecificFeatures_StdDevs()
        self.addTypeSpecificFeatures_HLV()
        self.addTypeSpecificFeatures_JetMoment()
        self.addCombinedFeatures_TypeVsType()
        self.addCombinedFeatures_SelectedTypes()
        
    #end init
    
    
    def dump(self):
        print("config_FeatureCalculator: Dumping features:")
        for iFeature in sorted(self.m_Feature_Names.iterkeys()):
            featName    = self.m_Feature_Names[iFeature]
            featType    = self.m_Feature_Types[iFeature]
            featDefault = self.m_Feature_Defaults[iFeature]
            print("\t" + string.ljust(featName, 50) + " of type " + featType + " defaults to " + string.rjust(str(featDefault), 10) )
            #print("\t" + featName + " of type " + featType + " defaults to " + str(featDefault) )
        #end for over features
        print("config_FeatureCalculator: Done dumping features")
    #end dump
    
    
    def addSingleVariables(self):
        Variables = []
        Variables += ["isPanTauCandidate"]
        Variables += ["RecoMode"]
        Variables += ["RecoMode_PanTau"]
        Variables += ["BDTValue_1p0n_vs_1p1n"]
        Variables += ["BDTValue_1p1n_vs_1pXn"]
        Variables += ["BDTValue_3p0n_vs_3pXn"]
        Variables += ["SumCharge"]
        Variables += ["AbsCharge"]
        for iVar in Variables:
            curDefVal = self.m_DefaultValues[self.m_VarTypeName_Basic]
            if iVar == "isPanTauCandidate":
                curDefVal = 0
            if iVar == "RecoMode" or iVar == "RecoMode_PanTau":
                curDefVal = -10
            self.addToFeatures(iVar, self.m_VarTypeName_Basic, "F", curDefVal)
        #end loop over variables
    #end def addSingleVariables
    
    
    def addMultiplicities(self):
        Variables = self.m_ConstituentTypes
        
        for iVar in Variables:
            featName = "N" + iVar + "Consts"
            self.addToFeatures(featName, self.m_VarTypeName_Basic, "F")
        #end loop over variables
        
        #add the inclusive one
        self.addToFeatures("N" + self.m_ConstituentTypeName_All     + "Consts", self.m_VarTypeName_Basic, "F")
        self.addToFeatures("N" + self.m_ConstituentTypeName_OutChrg + "Consts", self.m_VarTypeName_Basic, "F")
        self.addToFeatures("N" + self.m_ConstituentTypeName_OutNeut + "Consts", self.m_VarTypeName_Basic, "F")
        
    #end def addMultiplicities
    
    
    def addFourMomentum(self):
        Variables = []
        Variables += ["pt"]
        Variables += ["eta"]
        Variables += ["phi"]
        Variables += ["m"]
        
        Types = []
        Types += ["ProtoMomentumCore"]
        Types += ["ProtoMomentumWide"]
        
        for iVar in Variables:
            for iType in Types:
                curFeatName = iType + "_" + iVar
                curDefVal   = -4000.
                if iVar == "eta" or iVar == "phi":
                    curDefVal = -9.
                if iVar == "m":
                    curDefVal = -200.
                self.addToFeatures(curFeatName, self.m_VarTypeName_Basic, "F", curDefVal)
            #end loop types
        #end loop vars 
    #end def addFourMomentum
    
    def addTypeSpecificFeatures_PID(self):
        Variables = []
        Variables += ["BDTValues"]
        
        Sorts = []
        Sorts += ["BDTSort"]
        
        MaxNum = 4
        
        for iVar in Variables:
            for iSort in Sorts:
                for iNum in range(1, MaxNum):
                    featName = iVar + "_" + iSort + "_" + str(iNum)
                    self.addToFeatures_AllTypes(featName, self.m_VarTypeName_PID, "F")
                #end loop over num
            #end loop over sort
        #end loop over variables
    #end addTypeSpecificFeatures
    
    
    def addTypeSpecificFeatures_Shots(self):
        Variables = []
        Variables += ["nPhotons"]
        Variables += ["nShots"]
        Variables += ["SumShots_Et"]
        Variables += ["SumShots_Eta"]
        Variables += ["SumShots_Phi"]
        Variables += ["SumShots_M"]
        Variables += ["EtSumShotsOverConstEt"]
        
        Sorts = []
        Sorts += ["BDTSort"]
        
        MaxNum = 4
        
        for iVar in Variables:
            for iSort in Sorts:
                for iNum in range(1, MaxNum):
                    featName = iVar + "_" + iSort + "_" + str(iNum)
                    curDefVal = self.m_DefaultValues[self.m_VarTypeName_Shots]
                    if iVar == "Et":
                        curDefVal = -1000
                    if iVar == "Eta" or iVar == "Phi":
                        curDefVal = -8
                    if iVar == "M":
                        curDefVal = -200
                    
                    self.addToFeatures_FullName(self.m_ConstituentTypeName_Neutral + "_" + self.m_VarTypeName_Shots + "_" + featName, self.m_VarTypeName_Shots, "F", curDefVal)
                    self.addToFeatures_FullName(self.m_ConstituentTypeName_Pi0Neut + "_" + self.m_VarTypeName_Shots + "_" + featName, self.m_VarTypeName_Shots, "F", curDefVal)
                #end loop over iNum
            #end loop over sorts
        #end loop over variables
        
        Variables = []
        Variables += ["NShotsInSeed"]
        Variables += ["NPhotonsInSeed"]
        for iVar in Variables:
            featName = iVar
            curDefVal = self.m_DefaultValues[self.m_VarTypeName_Shots]
            
            self.addToFeatures_FullName(self.m_ConstituentTypeName_Neutral + "_" + self.m_VarTypeName_Shots + "_" + featName, self.m_VarTypeName_Shots, "F", curDefVal)
            self.addToFeatures_FullName(self.m_ConstituentTypeName_Pi0Neut + "_" + self.m_VarTypeName_Shots + "_" + featName, self.m_VarTypeName_Shots, "F", curDefVal)
        #end loop over variables
    #end addTypeSpecificFeatures_Shots
    
    def addTypeSpecificFeatures_Ratios(self):
        
        Variables_WithEnergyTypes = []
        Variables_WithEnergyTypes += ["EtOver"]
        Variables_WithEnergyTypes += ["1stEtOver"]
        Variables_WithEnergyTypes += ["1stBDTEtOver"]
        for iVar in Variables_WithEnergyTypes:
            self.addToFeatures_AllTypes_AllEnergyVariants(iVar, self.m_VarTypeName_Ratio, "F")
        #end loop over variables with energy types
        
        Variables = []
        Variables += ["1stEtOverTypeEt"]
        Variables += ["1stBDTEtOverTypeEt"]
        Variables += ["EFOsOverTotalEFOs"]
        for iVar in Variables:
            self.addToFeatures_AllTypes(iVar, self.m_VarTypeName_Ratio, "F")
        #end loop over variables with energy types
    #end addTypeSpecificFeatures_Ratios
    
    def addTypeSpecificFeatures_StdDevs(self):
        
        Variables_WithEnergyTypes = []
        Variables_WithEnergyTypes += ["Et_Wrt"]
        for iVar in Variables_WithEnergyTypes:
            self.addToFeatures_AllTypes_AllEnergyVariants(iVar, self.m_VarTypeName_StdDev, "F")
    #end addTypeSpecificFeatures_StdDevs
    
    
    def addTypeSpecificFeatures_HLV(self):
        Variables = []
        Variables += ["SumPt"]
        Variables += ["SumEta"]
        Variables += ["SumPhi"]
        Variables += ["SumM"]
        for iVar in Variables:
            curDefVal   = -4000.
            if iVar == "SumEta" or iVar == "SumPhi":
                curDefVal = -9.
            if iVar == "SumM":
                curDefVal = -200.
            self.addToFeatures_AllTypes(iVar, self.m_VarTypeName_HLV, "F", curDefVal)
            
            # also add the OutChrg and OutNeut
            featName = self.m_ConstituentTypeName_OutChrg + "_" + self.m_VarTypeName_HLV + "_" + iVar
            self.addToFeatures_FullName(featName, self.m_VarTypeName_HLV, "F", curDefVal)
            
            featName = self.m_ConstituentTypeName_OutNeut + "_" + self.m_VarTypeName_HLV + "_" + iVar
            self.addToFeatures_FullName(featName, self.m_VarTypeName_HLV, "F", curDefVal)
            
        #end loop over variables
        
        VariablesVec = []
        VariablesVec += ["Constituents_pt"]
        VariablesVec += ["Constituents_eta"]
        VariablesVec += ["Constituents_phi"]
        VariablesVec += ["Constituents_m"]
        
        VariablesVecSort = []
        VariablesVecSort += ["BDTSort"]
        
        for iSort in VariablesVecSort:
            for iVecVar in VariablesVec:
                featName = iSort + "_" + iVecVar
                curDefVal   = -4000.
                if iVecVar == "Constituents_eta" or iVecVar == "Constituents_phi":
                    curDefVal = -9.
                if iVecVar == "Constituents_m":
                    curDefVal = -200.
                self.addToFeatures_AllTypes(featName, self.m_VarTypeName_HLV, "V", curDefVal)
                
                # also add the OutChrg and OutNeut
                featName = self.m_ConstituentTypeName_OutChrg + "_" + self.m_VarTypeName_HLV + "_" + iSort + "_" + iVecVar
                self.addToFeatures_FullName(featName, self.m_VarTypeName_HLV, "V", curDefVal)
                
                featName = self.m_ConstituentTypeName_OutNeut + "_" + self.m_VarTypeName_HLV + "_" + iSort + "_" + iVecVar
                self.addToFeatures_FullName(featName, self.m_VarTypeName_HLV, "V", curDefVal)
                
        #end loop over variables
    #end addTypeSpecificFeatures_HLV
    
    def addTypeSpecificFeatures_JetMoment(self):
        Variables = []
        Variables += ["EtDR"]
        Variables += ["EtDRxTotalEt"]
        for iVar in Variables:
            self.addToFeatures_AllTypes(iVar, self.m_VarTypeName_JetMoment, "F")
        #end loop over variables
    #end addTypeSpecificFeatures_JetMoment
    
    def addCombinedFeatures_TypeVsType(self):
        Types = []
        Types += ["Charged"]
        Types += ["Neutral"]
        Types += ["Pi0Neut"]
        
        Variables = []
        Variables += [ ["Log1st", "EtOver", "Et"] ]
        Variables += [ ["Log", "EtOver", "Et"] ]
        Variables += [ ["Angle", "To", ""] ]
        Variables += [ ["DeltaR1st", "To1st", ""] ]
        Variables += [ ["Angle1st", "To1st", ""] ]
        for iType in Types:
            for jType in Types:
                
                #skip if types are equal
                if iType == jType:
                    continue
                
                for iVar in Variables:
                    featName = iVar[0] + iType + iVar[1] + jType + iVar[2]
                    #if iVar[2] != "-":
                        #featName = featName + iVar[2]
                    self.addToFeatures(featName, self.m_VarTypeName_Combined, "F")
                #end loop over variables
            #end loop over jType
        #end loop over iType
    #end addCombinedFeatures_TypeVsType
    
    
    def addCombinedFeatures_SelectedTypes(self):
        iTypes = []
        iTypes += ["Charged"]
        
        jTypes = []
        jTypes += ["Pi0Neut"]
        jTypes += ["Neutral"]
        
        Variables_WithEnergyTypes = []
        Variables_WithEnergyTypes += [ ["Mean", "Et_Wrt"] ]
        
        Variables = []
        Variables += [ ["InvMass", ""] ]
        for iType in iTypes:
            for jType in jTypes:
                
                for iVar in Variables:
                    featName = iVar[0] + iType + jType + iVar[1]
                    self.addToFeatures(featName, self.m_VarTypeName_Combined, "F")
                #end loop over variables
                
                for iVar in Variables_WithEnergyTypes:
                    featName = iVar[0] + iType + jType + iVar[1]
                    for iEnergyType in self.m_EnergyVariantsList:
                        featNameWithEnergy = featName + iEnergyType
                        self.addToFeatures(featNameWithEnergy, self.m_VarTypeName_Combined, "F")
                #end loop over variables for energy types
                
            #end loop over jTypes
        #end loop over iTypes
    #end addCombinedFeatures_SelectedTypes
        
#end class config_FeatureCalculator



