/*
  Copyright (C) 2002-2024 CERN for the benefit of the ATLAS collaboration
*/

#ifndef ISF_ISFEVENT_ISFTRUTHINCIDENT_H
#define ISF_ISFEVENT_ISFTRUTHINCIDENT_H

// stl includes
#include <vector>

// ISF includes
#include "ISF_Event/ISFParticleContainer.h"
#include "ISF_Event/ITruthIncident.h"
// DetectorDescription
#include "AtlasDetDescr/AtlasRegion.h" //enum

namespace ISF {
  class ISFParticle;

  /** @enum KillPrimary
      Basically only a boolean, which helps making the code more readable
  */
  enum KillPrimary {
    fPrimarySurvives = false,
    fKillsPrimary    = true
  };

  /** @class ISFTruthIncident

      Interface class for all truth incidents handled by the ISF.

      @author Elmar.Ritsch@cern.ch
  */

  class ISFTruthIncident : public ITruthIncident {
  public:
    ISFTruthIncident( ISF::ISFParticle &parent,
                      const ISFParticleVector &children,
                      int process,
                      AtlasDetDescr::AtlasRegion geoID,
                      ISF::KillPrimary killsPrimary = ISF::fPrimarySurvives,
                      const HepMC::FourVector *position = 0);
    virtual ~ISFTruthIncident();

    ISFTruthIncident( const ISFTruthIncident& ) = delete;
    ISFTruthIncident& operator= ( const ISFTruthIncident& ) = delete;

    /** Return HepMC position of the truth vertex */
    const HepMC::FourVector&  position() const override final;
    /** Return category of the physics process represented by the truth incident (eg hadronic, em, ..) */
    int                       physicsProcessCategory() const override final;
    /** Return specific physics process code of the truth incident (eg ionisation, bremsstrahlung, ..)*/
    int physicsProcessCode() const override final;

    /** Return p^2 of the parent particle */
    double                    parentP2() const override final;
    /** Return pT^2 of the parent particle */
    double                    parentPt2() const override final;
    /** Return Ekin of the parent particle */
    double                    parentEkin() const override final;
    /** Return the PDG Code of the parent particle */
    int                       parentPdgCode() const override final;
    /** Return the parent particle as a HepMC particle type
        (usually only called for particles that will enter the HepMC truth event) */
    HepMC::GenParticlePtr      parentParticle() override final;
    int      parentStatus() override final;
    /** Return the barcode of the parent particle */
    int  parentBarcode() override final; // TODO Remove this method
    /** Return the unique ID of the parent particle */
    int  parentUniqueID() override final;
    /** Return a boolean whether or not the parent particle survives the incident */
    bool                      parentSurvivesIncident() const override final;
    /** Return the parent particle after the TruthIncident vertex (and give
        it a new barcode) */
    HepMC::GenParticlePtr     parentParticleAfterIncident(int newBC) override final;

    /** Return p^2 of the i-th child particle */
    double                    childP2(unsigned short index) const override final;
    /** Return pT^2 of the i-th child particle */
    double                    childPt2(unsigned short index) const override final;
    /** Return Ekin of the i-th child particle */
    double                    childEkin(unsigned short index) const override final;
    /** Return the PDG Code of the i-th child particle */
    int                       childPdgCode(unsigned short index) const override final;
    /** Return the barcode of the i-th child particle (if defined as part of the TruthIncident) otherwise return 0 */
    int  childBarcode(unsigned short) const override final;
    /** Return the i-th child as a HepMC particle type and assign the given
        Barcode to the simulator particle (usually only called for particles that
        will enter the HepMC truth event) */
    HepMC::GenParticlePtr     childParticle(unsigned short index,
                                            int bc) override final;
    /** Update the id and particleLink properties of the parentAfterIncident (to be called after registerTruthIncident) */
    void updateParentAfterIncidentProperties();
    /** Update the id and particleLink properties of the child particles (to be called after registerTruthIncident) */
    void updateChildParticleProperties();

  private:
    ISFTruthIncident();

    /** return attached truth particle */
    inline HepMC::GenParticlePtr getHepMCTruthParticle( ISF::ISFParticle& particle ) const;

    /** convert ISFParticle to GenParticle and attach to ISFParticle's TruthBinding */
    inline HepMC::GenParticlePtr updateHepMCTruthParticle( ISF::ISFParticle& particle, ISF::ISFParticle* parent=nullptr ) const;

    ISF::ISFParticle&                  m_parent;
    const ISFParticleVector&           m_children;
    const int  m_process;
    const ISF::KillPrimary             m_killsPrimary;
    const HepMC::FourVector*           m_position;
  };

} // end of namespace

#endif // ISF_ISFEVENT_ISFTRUTHINCIDENT_H
