# Copyright (C) 2002-2024 CERN for the benefit of the ATLAS collaboration

# @file PyUtils.scripts.check_sg
# @purpose read a POOL file and dump the DataHeader's content
# @author Sebastien Binet
# @date February 2010

__doc__ = "read a POOL file and dump the DataHeader's content"
__author__ = "Sebastien Binet"


### imports -------------------------------------------------------------------
import PyUtils.acmdlib as acmdlib

@acmdlib.command(name='chk-sg')
@acmdlib.argument('files', nargs='+',
                  help='path to the POOL file(s) to analyze')
@acmdlib.argument('-o', '--output',
                  default=None,
                  help="""name of the output file which will contain the
                  informations gathered during checkSG processing.
                  These informations will be stored into a python-shelve or
                  an ASCII/py file (depending on the extension:
                  .pkl,.dat -> shelve; everything else -> ASCII/py)
                  """)
def main(args):
    """read a POOL file and dump the DataHeader's content

    ex:
     $ check-sg aod.pool.root
     $ check-sg /castor/foo.pool
     $ check-sg root://castoratlas//castor/foo.pool
     $ check-sg LFN:ttbar.pool
    """
    files = args.files
    if isinstance(files, str):
        files = [files]

    from collections import defaultdict
    from PyUtils.PoolFile import extract_items
    import os
    import sys

    for i,f in enumerate(files):
        files[i] = os.path.expandvars(os.path.expanduser(f))

    exitcode = 0
    for fname in files:
        try:
            print ("## checking [%s]..." % (fname,))

            item_list = defaultdict(list)
            for name, key in extract_items(fname, verbose=False, items_type='eventdata'):
                item_list[name].append(key)

            print ("="*80)
            print ("%40s%s%-40s" % ("Container type", " | ","StoreGate keys"))
            print ("%40s%s%-40s" % ("-"*40, "-+-", "-"*(40-3)))
            for name, sgkeys in sorted(item_list.items()):
                # decode b-strings to strings
                sgkeys = [(k.decode("utf-8") if not isinstance(k, str) else k) for k in sgkeys]
                print ("%40s%s%-40s" % (name, " | ", ', '.join(sorted(sgkeys))))
            print ("="*80)
            if args.output:
                outFileName = args.output
                outFileName = os.path.expanduser(outFileName)
                outFileName = os.path.expandvars(outFileName)
                print ("## saving report into [%s]..." % (outFileName,))
                if os.path.splitext(outFileName)[1] in ('.pkl', '.dat'):
                    import shelve
                    if os.path.exists(outFileName):
                        os.remove(outFileName)
                    with shelve.open(outFileName) as db:
                        db['eventdata_items'] = item_list
                else:
                    from pprint import pprint
                    with open( outFileName, 'w' ) as txt:
                        pprint( dict(item_list), txt )

        except Exception as e:
            print ("## Caught exception [%s] !!" % str(e.__class__))
            print ("## What:",e)
            print (sys.exc_info()[0])
            print (sys.exc_info()[1])
            import traceback
            traceback.print_exc()
            exitcode = 1
            pass

        if len(files) > 1:
            print ()
        pass # loop over fileNames
    
    print ("## Bye.")
    return exitcode

