/*
  Copyright (C) 2002-2025 CERN for the benefit of the ATLAS collaboration
*/

///////////////////////////////////////////////////////////////////
// Material.h, (c) ATLAS Detector software
///////////////////////////////////////////////////////////////////

#ifndef TRKGEOMETRY_BINNEDMATERIAL_H
#define TRKGEOMETRY_BINNEDMATERIAL_H

#include "TrkDetDescrUtils/CompactBinnedArray.h"
#include "TrkDetDescrUtils/CompactBinnedArray1D.h"
#include "TrkDetDescrUtils/CompactBinnedArray2D.h"
#include "TrkGeometry/Material.h"

#include <memory>
#include <vector>


namespace Trk {

typedef std::pair<std::shared_ptr<Material>, int> IdentifiedMaterial;

class BinUtility;

/** @class BinnedMaterial
 for description of non-homogenous dense volumes
 @author sarka.todorova@cern.ch
*/
class BinnedMaterial final: public Material {
 public:
  using binsPtr_t = std::unique_ptr<const CompactBinnedArray<const IdentifiedMaterial> >;

  /** Default Constructors */
  BinnedMaterial() = default;
  BinnedMaterial& operator=(BinnedMaterial&&) = default;
  BinnedMaterial(BinnedMaterial&&) = default;
  virtual ~BinnedMaterial() = default;

  //These can not be defaulted due to unique_ptr.
  BinnedMaterial(const BinnedMaterial& amc);
  BinnedMaterial& operator=(const BinnedMaterial& amc);

  /** Constructor with averaged material and binning in 1D*/
  BinnedMaterial(const Material& mat, const BinUtility& bu,
                 const std::vector<size_t>& index,
                 const std::vector<IdentifiedMaterial>& detailedMat);

  /** Constructor with averaged material and binning in 2D*/
  BinnedMaterial(const Material& mat, const BinUtility& bu,
                 const std::vector<Trk::BinUtility>& bVec,
                 const std::vector<std::vector<size_t> >& index,
                 const std::vector<IdentifiedMaterial>& detailedMat);


  /** access to layer bin utility */
  const Trk::BinUtility* layerBinUtility(const Amg::Vector3D& position) const;

  /** layer bin */
  size_t layerBin(const Amg::Vector3D& position) const;

  /** access to material/id per bin */
  const IdentifiedMaterial* material(const Amg::Vector3D& position) const;
  const IdentifiedMaterial* materialNext(const Amg::Vector3D& pos,
                                         const Amg::Vector3D& dir,
                                         bool layOnly) const;

 private:
  //utility to return pointers to the elements of m_matVec
  std::vector<const Trk::IdentifiedMaterial*> ptrs() const;

  //The vector that actual owns the IdentifiedMaterial elements
  std::vector<IdentifiedMaterial> m_matVec;
  // Binned arrays of IdentifiedMaterial elements
  binsPtr_t m_matBins;
};
}  // namespace Trk

inline const Trk::BinUtility* Trk::BinnedMaterial::layerBinUtility(
    const Amg::Vector3D& position) const {
  return m_matBins ? m_matBins->layerBinUtility(position) : nullptr;
}

inline size_t Trk::BinnedMaterial::layerBin(
    const Amg::Vector3D& position) const {
  return m_matBins ? m_matBins->layerBin(position) : 0;
}

#endif
