/*
  Copyright (C) 2002-2025 CERN for the benefit of the ATLAS collaboration
*/

#ifndef JIVEXML_TRACKRETRIEVER_H
#define JIVEXML_TRACKRETRIEVER_H

#include "GaudiKernel/ToolHandle.h"
#include "StoreGate/ReadCondHandleKey.h"
#include "JiveXML/IDataRetriever.h"
#include "AthenaBaseComps/AthAlgTool.h"
#include "GaudiKernel/ToolHandle.h"
#include "TrkToolInterfaces/ITrackSummaryTool.h"
#include "TrkToolInterfaces/IResidualPullCalculator.h"
#include "TrkTrack/TrackCollection.h"
#include "TrkTruthData/TrackTruthCollection.h"

//Forward declarations
class AtlasDetectorID;

namespace JiveXML{

  /**
   * @class TrackRetriever
   * @brief Retrieves all @c Trk::TrackCollection objects
   *
   *  - @b Properties
   *    - <em> PriorityTrackCollection</em><tt> = 'ExtendedTracks' </tt>:  @copydoc m_PriorityTrackCollection
   *    - <em> OtherTrackCollections</em><tt> = [ ] </tt>: @copydoc m_OtherTrackCollections
   *    - <em> TrackTruthCollections</em><tt> = [ ] </tt>: @copydoc m_TrackTruthCollections
   *    - <em> DoWriteHLT</em><tt> = False </tt>: @copydoc m_doWriteHLT
   *    - <em> DoWriteResiduals </em><tt> = True </tt>: @copydoc m_doWriteResiduals
   *    - <em> ResidualPullCalculator</em><tt> = PublicToolHandle('Trk::ResidualPullCalculator/ResidualPullCalculator')</tt>: @copydoc m_residualPullCalculator
   *    - <em> DoHitsSorting</em><tt> = True </tt>: @copydoc m_doHitsSorting
   *    - <em> DoHitsDetails</em><tt> = True </tt>: @copydoc m_doHitsDetails
   *
   *  - @b Retrieved @b Data
   *    - @e id : internal track ID counter
   *    - @e chi2 : @f$\chi^2@f$ of track fit
   *    - @e numDoF : number of degrees of freedom for the fit
   *    - @e trackAuthor : track author as defined in TrkTrack/TrackInfo.h
   *    - @e barcode : associated truth track barcode
   *    - @e numHits : number of hits associated to that track
   *    - @e covMatrix : entries of the covariance matrix
   *    - @e pt : transverse momentum
   *    - @e d0 : transverse impact parameter
   *    - @e z0 : longitudinal impact parameter
   *    - @e phi0 : initial phi direction at vertex
   *    - @e cotTheta : initial @f$\cos(\theta)@f$ at vertex
   *    - @e numPolyline : number of polyline points
   *    - <em> polylineX, polylineY, polylineZ </em>: polyline coordinates
   *    - @e numTsos : number of associated hits (Track State on Surface)
   *    - @e isOutlier : wether the hit is classified as outlier
   *    - @e driftSign : the drift sign of the hit
   *    - @e tsosDetType : detector type (one of 'PIX','SIL','TRT','MDT','CSC','RPC','TGC','unident')
   *    - @e tsosResLoc1 : local first coordinate of residual
   *    - @e tsosResLoc2 : local second coordinate of residual
   *    - @e tsosPullLoc1 : local first coordinate of pull
   *    - @e tsosPullLoc2 : local second coordinate of pull
   *    .
   */

  class TrackRetriever : virtual public IDataRetriever,
			 public AthAlgTool {

  public:

    /// Standard Constructor
    TrackRetriever(const std::string& type,const std::string& name,const IInterface* parent);

    virtual StatusCode retrieve(ToolHandle<IFormatTool> &FormatTool);
    /// Puts the variables into a DataMap
    const DataMap getData(const TrackCollection* trackCollection, const std::string &collectionName);
    /// Gets the StoreGate keys for the desired containers
    const std::vector<std::string> getKeys();
    /// Return the name of the data type that is generated by this retriever
    virtual std::string dataTypeName() const { return m_typeName; };

    ///Default AthAlgTool methods
    StatusCode initialize();

  private:
    ///The data type that is generated by this retriever
    const std::string m_typeName = "Track";

    Gaudi::Property<std::string> m_priorityKey {this,"PriorityTrackCollection","CombinedInDetTracks", "First track collections to retrieve, shown as default in Atlantis"};
    Gaudi::Property<std::vector<std::string>> m_otherKeys {this, "OtherTrackCollections", {}, "Track collections to retrieve, all if empty"};
    Gaudi::Property<bool> m_doWriteHLT {this, "DoWriteHLT", false, "Write out other collections that have HLT in the name"};
    Gaudi::Property<bool> m_doWriteAllCollections {this, "DoWriteAllCollections", false, "Write out all Track collections"};   
    Gaudi::Property<std::string> m_TrackTruthCollection {this, "TrackTruthColName", "TrackTruthCollection", "Track collection from which to retrieve the truth associations for the priority track collection"};
    Gaudi::Property<bool> m_doWriteResiduals {this,"DoWriteResiduals", true,"Whether to write TrackResiduals"};
    Gaudi::Property<bool> m_doHitsSorting {this, "DoHitsSorting", false, "Whether to perform hits TrackStateOnSurfaces (TSoS) sorting"};
    Gaudi::Property<bool> m_doHitsDetails {this, "DoHitsDetails", true, "Whether to write hits TrackStateOnSurfaces (TSoS) details"};

    ToolHandle<Trk::IResidualPullCalculator> m_residualPullCalculator {this, "ResidualPullCalculator","Trk::ResidualPullCalculator/ResidualPullCalculator", "ToolHandle to ResidualPullCaclulator" };
    ToolHandle< Trk::ITrackSummaryTool > m_trackSumTool{this, "TrackSummaryTool", "Trk::TrackSummaryTool/InDetTrackSummaryTool", "Track Summary Tool for number of Pix/SCT/TRT hits"};

    /// Used to find out the corresponding sub-det from ROT->identify().
    const AtlasDetectorID* m_idHelper{};

  };

}
#endif
