# Copyright (C) 2002-2024 CERN for the benefit of the ATLAS collaboration

from AthenaConfiguration.ComponentFactory import CompFactory
from AthenaConfiguration.ComponentAccumulator import ComponentAccumulator
from AthenaConfiguration.AccumulatorCache import AccumulatorCache
from TrigEDMConfig.TriggerEDM import recordable


@AccumulatorCache
def SPCounterRecoAlgCfg(flags):
    acc = ComponentAccumulator()
    from TrigMinBias.TrigMinBiasMonitoring import SpCountMonitoring
    alg = CompFactory.TrigCountSpacePoints(SpacePointsKey=recordable("HLT_SpacePointCounts"),
                                           MonTool=SpCountMonitoring(flags))
    alg.ModulesToSkip=[403] # see ATLHI-569
    acc.addEventAlgo(alg)
    return acc


def TrackCounterHypoAlgCfg(flags):
    """ Configuration for MinBias track counting algorithm """

    acc = ComponentAccumulator()
    from TrigMinBias.TrigMinBiasMonitoring import TrackCountMonitoring
    alg = CompFactory.TrackCountHypoAlg(name='TrackCountHypoAlg',
                                        tracksKey=recordable(flags.Trigger.InDetTracking.minBias.tracks_IDTrig),
                                        trackCountKey=recordable("HLT_TrackCount"))
    alg.MonTool = TrackCountMonitoring(flags, alg)  # monitoring tool configures itself using config of the hypo alg
    acc.addEventAlgo(alg)
    return acc


def PixelTrackCounterHypoAlgCfg(flags):
    """ Configuration for MinBias *Pixel* track counting algorithm """

    acc = ComponentAccumulator()
    from TrigMinBias.TrigMinBiasMonitoring import TrackCountMonitoring
    alg = CompFactory.TrackCountHypoAlg(name='PixelTrackCountHypoAlg',
                                        tracksKey=recordable(flags.Trigger.InDetTracking.minBiasPixel.tracks_IDTrig),
                                        trackCountKey=recordable("HLT_PixelTrackCount"))
    alg.MonTool = TrackCountMonitoring(flags, alg)  # monitoring tool configures itself using config of the hypo alg
    acc.addEventAlgo(alg)
    return acc


if __name__ == '__main__':
    from AthenaConfiguration.AllConfigFlags import initConfigFlags
    from AthenaConfiguration.TestDefaults import defaultTestFiles

    flags = initConfigFlags()
    flags.Input.Files = defaultTestFiles.RAW_RUN2  # or ESD or AOD or ...
    flags.lock()

    acc = ComponentAccumulator()
    acc.merge(SPCounterRecoAlgCfg(flags))
    acc.merge(TrackCounterHypoAlgCfg(flags))
    acc.merge(PixelTrackCounterHypoAlgCfg(flags))

    acc.printConfig(withDetails=True, summariseProps=True)
    acc.wasMerged()
